#ifndef TM_BASE_TYPES_H_
#define TM_BASE_TYPES_H_

#include <cuda_runtime.h>
#include <float.h>

#define BLOCK_SIZE 16
#define N_THREADS_FOR_REDUCE 512
#define IMG_T_MAX FLT_MIN
#define SKIP_COLOR 255	// white

#define EXPAND_TO_BLOCK_SIZE(x) (x % BLOCK_SIZE == 0 ? x : x - x % BLOCK_SIZE + BLOCK_SIZE)
#define EXPAND_TO_FULL_BLOCK_SIZE(x) (x % (BLOCK_SIZE * BLOCK_SIZE) == 0 ? x : x - x % (BLOCK_SIZE * BLOCK_SIZE) + (BLOCK_SIZE * BLOCK_SIZE))
#define CV_TIME(t) (((double) getTickCount() - t) / getTickFrequency())

namespace recognition {

template < typename T >
class IBuilder {
public:
	virtual ~IBuilder() {}
	virtual T build() = 0;
};

typedef float img_t;

// Matrices are stored in row-major order:
// M(row, col) = *(M.elements + row * M.width + col)
template <typename T>
struct Matrix {
	__host__ __device__ Matrix(int _height = 0, int _width = 0) : height(_height), width(_width), stride(_width), elements(NULL) {}

    int width;
    int height;
    int stride;
    T* elements;
};

typedef Matrix<unsigned char>	Matrix2uchar;
typedef Matrix<img_t>			Matrix2f;
typedef Matrix<double>			Matrix2d;

// List of coordinates denotes valuable elements
struct PointList {
	__host__ __device__ PointList(int _x, int _y) : x(_x), y(_y), next(NULL) {}
	__host__ __device__ PointList(int _x, int _y, PointList *_next) : x(_x), y(_y), next(_next) {}

	// Position of element in original image
	int x;
	int y;

	// Next element
	PointList *next;

	__host__ __device__ PointList ofSelf() {
		return PointList(-1, -1, this);
	}
};

struct PPointList {
	__host__ PPointList(PointList *_list, PPointList *_next): list(_list), next(_next) {}

	PointList *list;	// pointer to device memory, allocated by cudaMalloc
	PPointList *next;

	__host__ PPointList ofSelf() {
		return PPointList(NULL, this);
	}

	__host__ void _cu_freeList() {
		PPointList *cur = this;
		while (cur != NULL) {
			cudaFree(cur->list);
			PPointList *next = cur->next;
			free(cur);
			cur = next;
		}
	}
};

// Sparse image (other words, part of original image)
struct SparseImage {
	__host__ __device__ SparseImage(int _height, int _stride, img_t *_elements, PointList *_list, int _n_elements):
						height(_height), stride(_stride), elements(_elements), list(_list), n_elements(_n_elements) {}

	int height;			// image height
	int stride;			// original image width
	img_t *elements;	// elements of original image
	PointList *list;	// list of valuable elements
	int n_elements;		// total number of elements in image

	__host__ void freeList() {
		while (list != NULL) {
			PointList *cur = list;
			list = list->next;
			free(cur);
		}
	}

	__host__ void freeResources() {
		freeList();
		free(elements);
	}

	__device__ void _cu_freeList() {
		while (list != NULL) {
			PointList *cur = list;
			list = list->next;
			free(cur);
		}
	}

	__device__ void _cu_freeResources() {
		_cu_freeList();
		free(elements);
	}
};

struct CorrelationResult {
	CorrelationResult(): deviation(0), time(0) {}

	float deviation;
	float time;
};

class CorrelationPoint {
public:
	int x;
	int y;
	img_t corr;

	__device__ volatile CorrelationPoint& operator += (volatile const CorrelationPoint &rhs) volatile;

	__host__ __device__ CorrelationPoint& operator = (CorrelationPoint &rhs);
};

struct Record {
	float p;
	float time;
	float deviation;
	Record *next;
};

}	// namespace recognition

#endif	// TM_BASE_TYPES_H_
