#ifndef TM_GPU_CIRATEFI_H_
#define TM_GPU_CIRATEFI_H_

#include <cuda_runtime.h>
#include <cv.h>

namespace recognition {
namespace gpu {

class Ciratefi {
public:
	class Builder : public IBuilder<Ciratefi> {
	public:
		Builder& I1(cv::Mat &image) {
			this->I1_ = image;
			return *this;
		}

		Builder& I2(cv::Mat &image) {
			this->I2_ = image;
			return *this;
		}

		Builder& BC_thresholds(BCThresholds thresholds) {
			this->bc_thresholds_ = thresholds;
			return *this;
		}

		Builder& thresholds(CiratefiThresholds thresholds) {
			this->thresholds_ = thresholds;
			return *this;
		}

		Builder& bounds(CiratefiBoundaries bounds) {
			this->bounds_ = bounds;
			return *this;
		}

		Builder& maximum_shared_memory(int maximum_shared_memory) {
			this->maximum_shared_memory_ = maximum_shared_memory;
			return *this;
		}

		Ciratefi build() {
			return Ciratefi(this);
		}

	private:
		cv::Mat I1_, I2_;
		BCThresholds bc_thresholds_;
		CiratefiThresholds thresholds_;
		CiratefiBoundaries bounds_;
		int maximum_shared_memory_;

		friend class Ciratefi;
	};

	static const log4cxx::LoggerPtr logger;
	static const log4cxx::LoggerPtr output;

	static std::vector<Point2D> MidpointLine(int x0, int y_0, int x1, int y_1, int line_incr);
	static std::vector<Point2D> MidpointCircle(int radius, int angle_incr);
	static void CirclePoints(int x, int y, std::vector<Point2D>& circle);

	~Ciratefi() {
		destroyQ();
//		clearQs();
		checkCudaErrors(cudaFreeHost(A_expanded_and_alligned_by_block_size_.elements));
		cudaFree(d_C_A_elements_);
		cudaFree(d_R_A_);
		cudaFree(d_A_elements_);
		cudaFree(d_C_A_);
		cudaFree(d_A_);
	}

	void CreateCircles();
	void CreateLines();

	cv::Mat C_Q();
	cv::Mat C_A();
	cv::Mat _cu_C_A();
	std::vector<FirstGradePoint> CisCorr_AQ(cv::Mat& CQ, cv::Mat& CA);

	cv::Mat R_Q();
	cv::Mat R_A(std::vector<FirstGradePoint> points);
	std::vector<SecondGradePoint> RasCorr_AQ(cv::Mat& RQ, cv::Mat& RA, std::vector<FirstGradePoint> points);

	std::vector<FirstGradePoint> _cu_CisCorr_AQ(cv::Mat& CQ);
	cv::Mat _cu_R_A__global__(std::vector<FirstGradePoint> first_grade_points);
	std::vector<SecondGradePoint> _cu_RasCorr_AQ(cv::Mat& RQ, std::vector<FirstGradePoint> first_grade_points);

	double* Q();
//	double* Q(double s);
	void destroyQ();
	void clearQs();
	CiratefiResult Tefi(std::vector<SecondGradePoint> points);

	double Cis(cv::Mat& I, Point2D center, int radius);
	double Ras(cv::Mat& I, Point2D start, int radius, double angle);

	CiratefiResult FindCiratefi();

	double corr(double* x, double* y, int npoints);

	// Fields accessors
	cv::Mat& I1() { return I1_; }
	cv::Mat& I2() { return I2_; }
	BCThresholds bc_thresholds() const { return bc_thresholds_; }
	CiratefiThresholds thresholds() const { return thresholds_; }
	CiratefiBoundaries bounds() const { return bounds_; }

	int n() const { return n_; }
	int l() const { return l_; }
	int m() const { return m_; }

private:
	Ciratefi(Builder* builder) {
		this->I1_ = builder->I1_;
		this->I2_ = builder->I2_;
		this->bc_thresholds_ = builder->bc_thresholds_;
		this->thresholds_ = builder->thresholds_;
		this->bounds_ = builder->bounds_;
		this->maximum_shared_memory_ = builder->maximum_shared_memory_;
		this->maximum_shared_memory_ = 0;

		init();
	}

	void init() {
		this->offset_ = 0;
		this->out_of_boundaries_value_ = 100.0;

		this->Q_ = NULL;

		// TODO: more severe check. Introduce private method for that
		// Adjust boundaries
		if (bounds_.s.step == 0) bounds_.s.step = 0.1;
		if (bounds_.r.step == 0) bounds_.r.step = 1;
		if (bounds_.phi.step == 0) bounds_.phi.step = 2 * M_PI;
		bounds_.r.min = (bounds_.r.min / bounds_.r.step) * bounds_.r.step;
		bounds_.r.max = (bounds_.r.max / bounds_.r.step) * bounds_.r.step;

		this->n_ = (bounds_.s.max - bounds_.s.min) / bounds_.s.step + 1;
		this->l_ = (bounds_.r.max - bounds_.r.min) / bounds_.r.step + 1;
		this->m_ = (bounds_.phi.max - bounds_.phi.min) / bounds_.phi.step + 1;
		this->max_lambda_ = bounds_.r.max * bounds_.s.max;

		A_expanded_and_alligned_by_block_size_.stride = A_expanded_and_alligned_by_block_size_.width = 0;
		A_expanded_and_alligned_by_block_size_.height = 0;
		A_expanded_and_alligned_by_block_size_.elements = NULL;

		circle_radius_with_max_points_ = 0;
		max_num_of_circle_points_ = 0;
		total_num_of_circle_points_ = 0;

		max_num_of_line_points_ = 0;
	}

	int RadiusToIdx(int radius) {
		return recognition::RadiusToIdx(radius, bounds_);
	}

	int IdxToRadius(int idx) {
		return recognition::IdxToRadius(idx, bounds_);
	}

	int AngleToIdx(double angle) {
		return recognition::AngleToIdx(angle, bounds_);
	}

	double IdxToAngle(int idx) {
		return recognition::IdxToAngle(idx, bounds_);
	}

	cv::Mat& A_expanded();
	Matrix2d A_expanded_and_alligned_by_block_size();

	cv::Mat _cu_C_A__shared__();
	cv::Mat _cu_C_A__global__();

	cv::Mat I1_, I2_;
	BCThresholds bc_thresholds_;
	CiratefiThresholds thresholds_;
	CiratefiBoundaries bounds_;
	int maximum_shared_memory_;

	int n_, l_, m_;
	int max_lambda_;
	int max_num_of_circle_points_;
	int max_num_of_line_points_;
	int circle_radius_with_max_points_;
	int total_num_of_circle_points_;

	cv::Mat A_expanded_;
	Matrix2d A_expanded_and_alligned_by_block_size_;

	int offset_;
	double out_of_boundaries_value_;

	std::map<int, std::vector<Point2D> > circle_points_;
	std::map<int, std::map<double, std::vector<Point2D> > > line_points_;

	double* Q_;
//	std::map<double, double*> Qs;

	Matrix2d* d_A_;
	double* d_A_elements_;
	Matrix2d* d_C_A_;
	double* d_C_A_elements_;
	double* d_R_A_;

	std::vector<std::vector<Point2D> > circles_;
	std::vector<std::vector<std::vector<Point2D> > > lines_;
};

}	// namespace gpu
}	// namespace recognition

#endif	// TM_GPU_CIRATEFI_H_
