#ifndef TM_GPU_FAST_MATCH_H_
#define TM_GPU_FAST_MATCH_H_

#include <cuda_runtime.h>
#include <cv.h>

namespace recognition {
namespace gpu {

class BestTransformation {
public:
	class Builder : public IBuilder<BestTransformation> {
//	class Builder {
	public:
		Builder& Template(cv::Mat &I1) {
			this->I1 = I1;
			return *this;
		}

		Builder& Image(cv::Mat &I2) {
			this->I2 = I2;
			return *this;
		}

		Builder& Bounds(NetParameters bounds_min, NetParameters bounds_max) {
			this->bounds_min = bounds_min;
			this->bounds_max = bounds_max;
			return *this;
		}

		Builder& Steps(NetParameters steps) {
			this->steps = steps;
			return *this;
		}

		Builder& Epsilon(double epsilon) {
			this->epsilon = epsilon;
			return *this;
		}

		Builder& Delta(double delta) {
			this->delta = delta;
			return *this;
		}

		Builder& PhotometricInvariance(double photometricInvariance) {
			this->photometricInvariance = photometricInvariance;
			return *this;
		}

		BestTransformation build() {
			return BestTransformation(this);
		}

	private:
		cv::Mat I1, I2;
		NetParameters bounds_min, bounds_max, steps;
		double epsilon, delta, photometricInvariance;

		friend class BestTransformation;
	};

	enum expansionType { RANDOM_EXPANSION, FULL_EXPANSION };

	static const log4cxx::LoggerPtr logger;
	static const log4cxx::LoggerPtr output;

	BestTransformationResult FindBestTransformation();

	void CreateListOfConfigs(cv::Mat& configs);

	void CreateList(int ntx_steps, int nty_steps, int nr_steps, int NR2_steps, int ns_steps,
			cv::Mat& tx_steps, cv::Mat& ty_steps, cv::Mat& r_steps, cv::Mat& s_steps, cv::Mat &configs);

	void Configs2Affine(cv::Mat& configs, cv::Mat& affines, cv::Mat& insiders, int &n_in_bounds);

	void FilterByBounds(cv::Mat& src, cv::Mat& dst, cv::Mat& filter);

	void EvaluateConfigsCuda(cv::Mat& I1, cv::Mat& I2, cv::Mat& configs_, cv::Mat& xs_, cv::Mat& ys_, int photometricInvariance, cv::Mat& distances_);
	void EvaluateConfigs(cv::Mat& I1, cv::Mat& I2, cv::Mat& configs_, cv::Mat& xs_, cv::Mat& ys_, int photometricInvariance, cv::Mat& distances_);

	void CreateAffineTransformation(cv::Mat& config, cv::Mat& transf);

	void GetGoodConfigsByDistance(cv::Mat& configs, double bestDist, double newDelta, cv::Mat& distances, cv::Mat& goodConfigs, int& tooHighPercentage);

	double GetThreshPerDelta(double delta);

	void ExpandConfigsRandom(cv::Mat& configs, int level, int npoints, double deltaFact, cv::Mat& expandedConfigs);

	void ExpandConfigsFull(cv::Mat& configs, int level, double deltaFact, cv::Mat& expandedConfigs);

//	void EvaluateConfigs2(int n_transformations, int n_points, double* pattern, double* image, double* cover, int w2, double c, int* x, int* y, double* correlations);

private:
	BestTransformation(Builder* builder) {
		this->I1 = builder->I1;
		this->I2 = builder->I2;
		this->bounds_min = builder->bounds_min;
		this->bounds_max = builder->bounds_max;
		this->steps = builder->steps;
		this->epsilon = builder->epsilon;
		this->delta = builder->delta;
		this->photometricInvariance = builder->photometricInvariance;
	}

	cv::Mat I1, I2;
	NetParameters bounds_min, bounds_max, steps;
	double epsilon, delta, photometricInvariance;
};

__global__ void _cu_findDistancesKernel(double* d_configs, int* d_xs_centered, int* d_ys_centered, double* d_valsI1,
		double *d_img2, _cu_findDistancesKernelArgs *args, double *d_distances);

}	// namespace gpu
}	// namespace recognition

#endif	// TM_GPU_FAST_MATCH_H_
