/**
 * According to the article "[35.13] How can I avoid linker errors with my template functions?":
 * http://www.parashift.com/c++-faq-lite/separate-template-fn-defn-from-decl.html
 */

#include <base/types.cuh>
#include <base/methods.cuh>

namespace recognition {

Matrix2f NormalizeImage(Matrix2uchar m_image, int *stoch_vect, const unsigned v_length, float skip, img_t &sum, unsigned &n_elements) {
	// Fill stochastic vector
	for (int i = 0; i < v_length; i++) {
		float r = rand() / (double)RAND_MAX;
		stoch_vect[i] = r < skip ? 1 : 0;
	}

	int size = m_image.height * m_image.width;

	img_t Avr = 0, Norm = 0;
	unsigned stoch_index = 0;
	n_elements = 0;
	for (int i = 0; i < size; i++) {
		if (stoch_vect[stoch_index++ % v_length]) {
			continue;
		}

		img_t val = (img_t) m_image.elements[i];
		Avr += val;
		Norm += val * val;
		n_elements++;
	}

	Avr /= n_elements;
	Norm /= n_elements;
	Norm -= Avr * Avr;
	Norm = sqrt(Norm);
	if (Norm == 0)
		Norm = 1;

	Matrix2f res;
	res.height = m_image.height;
	res.width = m_image.width;
	res.stride = m_image.stride;
	res.elements = (img_t *) malloc(sizeof(img_t) * size);

	sum = 0;
	stoch_index = 0;
	for (int i = 0; i < size; i++) {
		if (stoch_vect[stoch_index++ % v_length]) {
			res.elements[i] = 0;
			continue;
		}

		img_t val = ((img_t) m_image.elements[i] - Avr) / Norm;
		res.elements[i] = val;
		sum += val;
	}

	return res;
}


void NormalizeImage(SparseImage &img, img_t &sum) {
	PointList start = img.list->ofSelf();
	PointList *list = &start;
	img_t Avr = 0;
	img_t Norm = 0;

	while ((list = list->next) != NULL) {
		img_t val = img.elements[list->y * img.stride + list->x];
		Avr += val;
		Norm += val * val;
	}

	Avr /= img.n_elements;
	Norm /= img.n_elements;
	Norm -= Avr * Avr;
	Norm = Norm != 0 ? sqrt(Norm) : 1;

	sum = 0;
	list = &start;
	while ((list = list->next) != NULL) {
		int idx = list->y * img.stride + list->x;
		img_t val = (img.elements[idx] - Avr) / Norm;
		img.elements[idx] = val;
		sum += val;
	}
}

template< typename T >
SparseImage GetSparseImage(Matrix<T> img, float p) {
	img_t *elements = (img_t*) malloc(sizeof(img_t) * img.height * img.width);
	SparseImage res(img.height, img.stride, elements, NULL, 0);

	int v_length = 10000;
	char stoch_vect[v_length];
	for (int i = 0; i < v_length; i++) {
		float r = rand() / (double)RAND_MAX;
		stoch_vect[i] = r < p ? 1 : 0;
	}
	unsigned int stoch_index = 0;

	PointList *list = NULL;
	for (int y = 0; y < img.height; ++y) {
		for (int x = 0; x < img.width; ++x) {
			int idx = y * img.stride + x;
			T val = img.elements[idx];
			elements[idx] = (img_t) val;
			if (val == SKIP_COLOR || stoch_vect[stoch_index++ % v_length])
				continue;

			if (res.list == NULL) {
				list = (PointList*) malloc(sizeof(PointList));
				list->x = x;
				list->y = y;
				list->next = NULL;
				res.list = list;
			} else {
				list->next = (PointList*) malloc(sizeof(PointList));
				list->next->x = x;
				list->next->y = y;
				list->next->next = NULL;
				list = list->next;
			}

			res.n_elements++;
		}
	}

	return res;
}

template SparseImage GetSparseImage(Matrix<unsigned char> img, float p);

template< typename R, typename T >
R* GetIntegralRepresentation(T* image, int height, int width, int power) {
	int x, y;
	R* res = (R*) malloc(sizeof(R) * height * width);

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			res[y * width + x] = pow(image[y * width + x], power)
							   - (y > 0 && x > 0 ? res[(y - 1) * width + x - 1] : 0)
							   + (y > 0 ? res[(y - 1) * width + x] : 0)
							   + (x > 0 ? res[y * width + x - 1] : 0);
		}
	}

	return res;
}

template< typename R, typename T >
R* GetIntegralRepr(T* image, int height, int width) {
	return GetIntegralRepresentation<R, T>(image, height, width, 1);
}

template< typename R, typename T >
R* GetIntegralReprSquared(T* image, int height, int width)  {
	return GetIntegralRepresentation<R, T>(image, height, width, 2);
}

template< typename T >
__host__ __device__ T EvalIntegralReprSum(T* image, int stride, int top, int left, int height, int width) {
	int idxA, idxB, idxC, idxD;
	T camA, camB, camC, camD;
	idxA = (top - 1) * stride + (left - 1);
	idxB = (top - 1) * stride + (left - 1) + width;
	idxC = (top - 1 + height) * stride + (left - 1) + width;
	idxD = (top - 1 + height) * stride + (left - 1);
	camA = top > 0 && left > 0 ? image[idxA] : 0;
	camB = top > 0 ? image[idxB] : 0;
	camC = image[idxC];
	camD = left > 0 ? image[idxD] : 0;

	return camA + camC - camB - camD;
}

template int* GetIntegralRepr(unsigned char *image, int height, int width);
template float* GetIntegralRepr(float *image, int height, int width);

template long* GetIntegralReprSquared(unsigned char *image, int height, int width);
template double* GetIntegralReprSquared(float *image, int height, int width);

template __host__ __device__ int EvalIntegralReprSum(int *image, int top, int left, int stride, int height, int width);
template __host__ __device__ long EvalIntegralReprSum(long *image, int top, int left, int stride, int height, int width);

}	// namespace recognition
