#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include <iostream>

#include <cuda_runtime.h>

#include <cv.h>
#include <highgui.h>
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/core/core.hpp"
#include "opencv2/highgui/highgui.hpp"

#include <log4cxx/logger.h>
#include <base/logging.h>
#include <base/types.cuh>

#include <gpu/fast_match_helper.cuh>
#include <gpu/fast_match.cuh>

#include <gpu/ciratefi_helper.cuh>
#include <gpu/ciratefi.cuh>

#include <debug/garbage.cuh>

#include <debug/simple_test.cuh>

using namespace cv;

namespace recognition {

void DEBUG_Recolor(Mat image, double color) {
	for (int y = 0; y < image.rows; ++y)
		for (int x = 0; x < image.cols; ++x)
			if (image.at<double>(y, x) == 0)
				image.at<double>(y, x) = color;
}

void DEBUG_DrawCircle(cv::Mat image, int x_offset, int y_offset, int radius, int angle_incr, double color, bool bold) {
	std::vector<Point2D> circle = recognition::MidpointCircle(radius, angle_incr);
	for (int i = 0; i < circle.size(); ++i) {
		int x, y;
		x = circle[i].x + x_offset;
		y = circle[i].y + y_offset;

		if (bold)
			DEBUG_DrawBoldPoint(image, y, x, color);
		else
			image.at<double>(y, x) = color;
	}
}

void DEBUG_DrawLine(Mat image, int x_offset, int y_offset, int x_start, int y_start, int x_end, int y_end, int line_incr, double color, bool bold) {
	std::vector<Point2D> line = recognition::MidpointLine(x_start, y_start, x_end, y_end, line_incr);
	for (int i = 0; i < line.size(); ++i) {
		int x, y;
		x = line[i].x + x_offset;
		y = line[i].y + y_offset;

		if (bold)
			DEBUG_DrawBoldPoint(image, y, x, color);
		else
			image.at<double>(y, x) = color;
	}
}

void DEBUG_AffineMatrix(Mat I1, Mat I2) {
	double s = 0.5;
	Mat I1_ = Mat::zeros(I1.rows * s, I1.cols * s, I1.type());

	double r = M_PI / 4;
	int tx = 100;
	int ty = 100;

	Mat R = (Mat_<double>(2, 2) << cos(r), -sin(r) , sin(r), cos(r));
	Mat S = (Mat_<double>(2, 2) << s, 0 , 0, s);
	Mat A = (Mat_<double>(2, 3) << 0, 0, tx , 0, 0, ty);

	Mat A_ = S * R;
	for (int y = 0; y < 2; ++y)
		for (int x = 0; x < 2; ++x)
			A.at<double>(y, x) = A_.at<double>(y, x);

	double a11, a12, a21, a22, a13, a23;
	a11 = A.at<double>(0, 0);
	a12 = A.at<double>(0, 1);
	a21 = A.at<double>(1, 0);
	a22 = A.at<double>(1, 1);

	// Detect sizes of transformed pattern
	Mat corners = (Mat_<int>(2, 4) <<	0, a11 * (I1.cols - 1), a11 * (I1.cols - 1) + a12 * (I1.rows - 1), a12 * (I1.rows - 1),
										0, a21 * (I1.cols - 1), a21 * (I1.cols - 1) + a22 * (I1.rows - 1), a22 * (I1.rows - 1));
	cout << "corners:\n" << corners;

	int xDim, yDim;	// sizes of transformed pattern
	int minX, minY, maxX, maxY;	// min, max of components x and y
	minX = minY = INT_MAX;
	maxX = maxY = INT_MIN;

	for (int i = 0; i < 4; ++i) {
		int x, y;
		x = corners.at<int>(0, i);
		y = corners.at<int>(1, i);

		if (x < minX) minX = x;
		if (x > maxX) maxX = x;
		if (y < minY) minY = y;
		if (y > maxY) maxY = y;
	}

	xDim = maxX - minX + 1;
	yDim = maxY - minY + 1;

	cout << "DIMS: " << xDim << " " << yDim << endl;
	I1_ = Mat::zeros(yDim, xDim, I1.type());

	Mat warp_mat = A;
	warp_mat.at<double>(0, 2) = -minX;
	warp_mat.at<double>(1, 2) = -minY;
	cout << "warp_mat:\n" << warp_mat << endl;

	/// Apply the Affine Transform just found to the src image
	warpAffine(I1, I1_, warp_mat, I1_.size());

	namedWindow("Pattern", CV_WINDOW_AUTOSIZE);
	imshow("Pattern", I1);

	namedWindow("Transformed pattern", CV_WINDOW_AUTOSIZE);
	imshow("Transformed pattern", I1_);

	a11 = s * cos(r); a12 = -s * sin(r);
	a21 = s * sin(r); a22 = s * cos(r);
	a13 = tx; a23 = ty;

	for (int y = 0; y < I1.rows; ++y)
		for (int x = 0; x < I1.cols; ++x) {
			int target_x, target_y;
			target_x = a11 * x + a12 * y + a13;
			target_y = a21 * x + a22 * y + a23;

			I2.at<double>(target_y, target_x) = I1.at<double>(y, x);
		}

	namedWindow("I2", CV_WINDOW_AUTOSIZE);
	imshow("I2", I2);

	for (;;) if (cv::waitKey(30) >= 0) exit(0);
}

void DEBUG_DrawBoldPoint(Mat image, int y, int x, double color) {
	image.at<double>(y, x) = color;
	image.at<double>(y, x - 1) = color;
	image.at<double>(y, x + 1) = color;
	image.at<double>(y - 1, x - 1) = color;
	image.at<double>(y - 1, x) = color;
	image.at<double>(y - 1, x + 1) = color;
	image.at<double>(y + 1, x - 1) = color;
	image.at<double>(y + 1, x) = color;
	image.at<double>(y + 1, x + 1) = color;
}

void DEBUG_NumberOfPoints(recognition::gpu::Ciratefi ciratefi) {
	CiratefiBoundaries bounds_ = ciratefi.bounds();
	Mat I1 = ciratefi.I1();
	Mat I2 = ciratefi.I2();

	printf("n*l=%d\n", ciratefi.n() * ciratefi.l());

	int npoints_of_lines = 0;
	double s_ = 2;
	int r_max = 128;
	double phi_step = 2 * M_PI / 36;
	int line_incr_ = 2;
	for (double phi = bounds_.phi.min; phi <= bounds_.phi.max; phi += phi_step) {
		npoints_of_lines += recognition::MidpointLine(0, 0,
				s_ * r_max * cos(phi), s_ * r_max * sin(phi), line_incr_).size();
	}
	printf("npoints_of_lines=%d sizeof(Point2D)=%zu\n", npoints_of_lines, sizeof(Point2D));

	std::vector<Point2D> circle = recognition::MidpointCircle(128, 2);
	printf("npoints_of_cirle=%d sizeof(Point2D)=%zu\n", circle.size(), sizeof(Point2D));

	int npoints_in_cirles = 0;
	for (int i = 0; i <= 128; ++i)
		npoints_in_cirles += recognition::MidpointCircle(i, 2).size();
	printf("npoints_in_cirles=%d\n", npoints_in_cirles);
	exit(0);
}

// ****************************************************************************
SecondGradePoint DEBUG_RunCiratefiCPU(recognition::gpu::Ciratefi ciratefi) {
	const log4cxx::LoggerPtr logger = recognition::gpu::Ciratefi::logger;

	LOG_INFO(logger, "Starting CIRATEFI...");
	LOG_INFO(logger, "Executing Cifi...");
	double ciratefi_time = (double)getTickCount();
	double cifi_time = (double)getTickCount();
	Mat CQ = ciratefi.C_Q();
	Mat CA = ciratefi.C_A();
	std::vector<FirstGradePoint> CisCorr_AQ = ciratefi.CisCorr_AQ(CQ, CA);
	LOG_INFO(logger, "Cifi completed with time: %.4fsec. First grade candidates: %zu", CV_TIME(cifi_time), CisCorr_AQ.size());

	LOG_INFO(logger, "Executing Rafi...");
	double rafi_time = (double)getTickCount();
	Mat RQ = ciratefi.R_Q();
	Mat RA = ciratefi.R_A(CisCorr_AQ);
	std::vector<SecondGradePoint> RasCorr_AQ = ciratefi.RasCorr_AQ(RQ, RA, CisCorr_AQ);
	LOG_INFO(logger, "Rafi completed with time: %.4fsec. First grade candidates: %zu", CV_TIME(rafi_time), RasCorr_AQ.size());

	LOG_INFO(logger, "Executing Tefi...");
	double tefi_time = (double)getTickCount();
	CiratefiResult ciratefi_result = ciratefi.Tefi(RasCorr_AQ);
	LOG_INFO(logger, "Tefi completed with time: %.4fsec", CV_TIME(tefi_time));
	LOG_INFO(logger, "CIRATEFI completed with time: %.4fsec", CV_TIME(ciratefi_time));

	if (ciratefi_result.points.size() == 0) {
		LOG_ERROR(logger, "No best match with given thresholds!");
		exit(0);
	}

	return ciratefi_result.best_match;
}

// ****************************************************************************
SecondGradePoint DEBUG_RunCiratefiGPU(recognition::gpu::Ciratefi ciratefi) {
	const log4cxx::LoggerPtr logger = recognition::gpu::Ciratefi::logger;

	LOG_INFO(logger, "Starting CIRATEFI...");
	LOG_INFO(logger, "Executing Cifi...");
	double ciratefi_time = (double)getTickCount();
	double cifi_time = (double)getTickCount();
	Mat CQ = ciratefi.C_Q();
	Mat CA = ciratefi._cu_C_A();
	std::vector<FirstGradePoint> CisCorr_AQ = ciratefi._cu_CisCorr_AQ(CQ);
	LOG_INFO(logger, "Cifi completed with time: %.4fsec. First grade candidates: %zu", CV_TIME(cifi_time), CisCorr_AQ.size());

	LOG_INFO(logger, "Executing Rafi...");
	double rafi_time = (double)getTickCount();
	Mat RQ = ciratefi.R_Q();
	Mat RA = ciratefi._cu_R_A__global__(CisCorr_AQ);
//	Mat RA1 = ciratefi.R_A(CisCorr_AQ);
//	for (int i = 0; i < 10; ++i) {
//		Point2D p = CisCorr_AQ[i].p;
//		for (int j = 0; j < ciratefi.m(); j++) {
//			printf("[%d: %f %f] ", j, RA.at<double>(p.x, p.y, j), RA1.at<double>(p.x, p.y, j));
//			if (j >= 11 && j <=14) {
//				printf("( %f ) ", IdxToAngle(j, bounds_));
//			}
//		}
//		printf("\n");
//	}
//	printf("\n");

	std::vector<SecondGradePoint> RasCorr_AQ = ciratefi._cu_RasCorr_AQ(RQ, CisCorr_AQ);
	LOG_INFO(logger, "Rafi completed with time: %.4fsec. First grade candidates: %zu", CV_TIME(rafi_time), RasCorr_AQ.size());

	LOG_INFO(logger, "Executing Tefi...");
	double tefi_time = (double)getTickCount();
	CiratefiResult ciratefi_result = ciratefi.Tefi(RasCorr_AQ);
	LOG_INFO(logger, "Tefi completed with time: %.4fsec", CV_TIME(tefi_time));
	LOG_INFO(logger, "CIRATEFI completed with time: %.4fsec", CV_TIME(ciratefi_time));

	if (ciratefi_result.points.size() == 0) {
		LOG_ERROR(logger, "No best match with given thresholds!");
		exit(0);
	}

	return ciratefi_result.best_match;
}

// ****************************************************************************
void DEBUG_RunCiratefi(recognition::gpu::Ciratefi ciratefi) {
	// Run on CPU
//	SecondGradePoint best_match = DEBUG_RunCiratefiCPU(ciratefi);

	// Run on GPU
	SecondGradePoint best_match = DEBUG_RunCiratefiGPU(ciratefi);

	CiratefiBoundaries bounds_ = ciratefi.bounds();
	Mat I1 = ciratefi.I1();
	Mat I2 = ciratefi.I2();

	// Draw first grade points on image
//	double color1 = 0;
//	for (int i = 0; i < CisCorr_AQ.size(); ++i) {
//		I2.at<double>(CisCorr_AQ[i].p.y, CisCorr_AQ[i].p.x) = color1;
//	}

	// Draw second grade points on image
//	double color2 = 1;
//	for (int i = 0; i < RasCorr_AQ.size(); ++i) {
//		DEBUG_DrawBoldPoint(I2, RasCorr_AQ[i].p.y, RasCorr_AQ[i].p.x, color2);
//	}

	int w1, h1, half_w1, half_h1;
	w1 = I1.cols;
	h1 = I1.rows;
	half_w1 = w1 / 2;
	half_h1 = h1 / 2;

	double s, phi;
	s = bounds_.s.min + best_match.scale_idx * bounds_.s.step;
	phi = bounds_.phi.min + best_match.angle_idx * bounds_.phi.step;

	double a11, a12, a21, a22, a13, a23;
	// -phi Counterclockwise
	a11 = s * cos(-phi); a12 = -s * sin(-phi);
	a21 = s * sin(-phi); a22 = s * cos(-phi);
	a13 = best_match.p.x; a23 = best_match.p.y;

//	Mat I2_copy = I2.clone();

	// Drow I1 over I2
	for (int y = -half_h1; y < h1 - half_h1; ++y) {
		for (int x = -half_w1; x < w1 - half_w1; ++x) {
			int target_x, target_y;
			target_x = a11 * x + a12 * y + a13;
			target_y = a21 * x + a22 * y + a23;

			I2.at<double>(target_y, target_x) = I1.at<double>(y + half_h1, x + half_w1) * 0.8;
		}
	}

//	Mat I1_circ = I1.clone();
//	for (int r1 = 1; r1 < min(I1_circ.cols / 2 - 1, I1_circ.rows / 2 - 1); r1 += 2) {
//		DEBUG_DrawCircle(I1_circ, I1_circ.cols / 2, I1_circ.rows / 2, r1, 1, 0, false);
//	}
//	imshow("I1_circ", I1_circ);
//	printf("ok");
//
//	Mat I1_lines = I1.clone();
//	double radius1 = min(I1_lines.cols / 2 - 1, I1_lines.rows / 2 - 1);
//	for (double phi1 = bounds_.phi.min; phi1 < bounds_.phi.max; phi1 += bounds_.phi.step * 2) {
//		DEBUG_DrawLine(I1_lines, I1_lines.cols / 2, I1_lines.rows / 2, 0, 0, radius1 * cos(phi1), radius1 * sin(phi1), 1, 0, false);
//	}
//	imshow("I1_lines", I1_lines);
//	printf("ok");

	// Drow circle and radius-vector
	int radius = s * min(half_w1, half_h1);
	int step = 1;
	DEBUG_DrawCircle(I2, best_match.p.x, best_match.p.y, radius, step, 0, false);
	DEBUG_DrawLine(I2, best_match.p.x, best_match.p.y, 0, 0, radius, 0, 2, 0, true);
	DEBUG_DrawLine(I2, best_match.p.x, best_match.p.y, 0, 0, radius * cos(-phi), radius * sin(-phi), step, 0, false);

//	DEBUG_DrawBoldPoint(I2, best_match.p.y, best_match.p.x, 0);

//	int resize_scale = 2;
	int resize_scale = 1;
	resize(I2, I2, Size(), resize_scale, resize_scale);
	imshow("I2", I2);
	imshow("I1", I1);

	ciratefi.~Ciratefi();
	for (;;) if (cv::waitKey(30) >= 0) exit(0);

//	resize(I2_copy, I2_copy, Size(), resize_scale, resize_scale);
//	imshow("I2_copy", I2_copy);


//	DEBUG_NumberOfPoints(ciratefi);

//	DEBUG_AffineMatrix(I1, I2);

//	// Draw a test circle in (100, 100) of radius 50 with step 5
//	DEBUG_DrawCircle(I2, 100, 100, 50, 5);
//
//	// Draw a test line between points (200, 200) and (350, 150) with step 2
//	DEBUG_DrawLine(I2, 200, 200, 0, 0, 150, -50, 2);
//
//	imshow("I2", I2);
//	for (;;) if (cv::waitKey(30) >= 0) exit(0);
}

}	// namespace recognition
