#include <stdio.h>
#include <base/types.cuh>
#include <gpu/cu_methods.cuh>

namespace recognition {
namespace gpu {

// Get a matrix element
__device__ img_t _cu_getElement(const Matrix2f A, int row, int col) {
	return A.elements[row * A.stride + col];
}

// Set a matrix element
__device__ void _cu_setElement(Matrix2f A, int row, int col, img_t value) {
	A.elements[row * A.stride + col] = value;
}

// Get the (width x height) sub-matrix Asub of A that is
// located col sub-matrices to the right and row sub-matrices down
// from the upper-left corner of A (all sub-matrices are aligned
// by BLOCK_SIZE both dimensions)
__device__ Matrix2f _cu_getSubMatrix(Matrix2f *A, int row, int col) {
	Matrix2f Asub;
	Asub.width = Asub.height = -1;	// undefined and not used
	Asub.stride = A->stride;
	Asub.elements = &A->elements[A->stride * BLOCK_SIZE * row + BLOCK_SIZE * col];
	return Asub;
}

__host__ void _cu_catch(cudaError_t error, const char *message) {
	if (error != cudaSuccess) {
		fprintf(stderr, "%s (error code: %s)!\n", message, cudaGetErrorString(error));
		exit(EXIT_FAILURE);
	}
}

template <unsigned int blockSize>
__device__ void _cu_warpReduce(volatile CorrelationPoint *sdata, unsigned int tid) {
	if (blockSize >= 64) sdata[tid] += sdata[tid + 32];
	if (blockSize >= 32) sdata[tid] += sdata[tid + 16];
	if (blockSize >= 16) sdata[tid] += sdata[tid + 8];
	if (blockSize >= 8) sdata[tid] += sdata[tid + 4];
	if (blockSize >= 4) sdata[tid] += sdata[tid + 2];
	if (blockSize >= 2) sdata[tid] += sdata[tid + 1];
}

template <unsigned int blockSize>
__global__ void _cu_reduceMaxKernel(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n) {
//	extern __shared__ CorrCoordinates sdata[];	// in case of dynamic allocation
	__shared__ CorrelationPoint sdata[blockSize];
	unsigned int tid = threadIdx.x;
	unsigned int i = blockIdx.x * (blockSize * 2) + tid;
	unsigned int gridSize = blockSize * 2 * gridDim.x;

	// Initialize first element
	sdata[tid].x = sdata[tid].y = -1;
	sdata[tid].corr = -IMG_T_MAX;

	while (i < n) {
		sdata[tid] += g_idata[i];
		sdata[tid] += g_idata[i + blockSize];
		i += gridSize;
	}
	__syncthreads();

	if (blockSize >= 512) { if (tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
	if (blockSize >= 256) { if (tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
	if (blockSize >= 128) { if (tid < 64) { sdata[tid] += sdata[tid + 64]; } __syncthreads(); }

	if (tid < 32) _cu_warpReduce<blockSize>(sdata, tid);
	if (tid == 0) g_odata[blockIdx.x] = sdata[0];
}

template __global__ void _cu_reduceMaxKernel<1024>(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n);
template __global__ void _cu_reduceMaxKernel<512>(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n);
template __global__ void _cu_reduceMaxKernel<256>(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n);
template __global__ void _cu_reduceMaxKernel<128>(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n);
template __global__ void _cu_reduceMaxKernel<64>(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n);
template __global__ void _cu_reduceMaxKernel<32>(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n);
template __global__ void _cu_reduceMaxKernel<16>(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n);
template __global__ void _cu_reduceMaxKernel<8>(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n);
template __global__ void _cu_reduceMaxKernel<4>(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n);
template __global__ void _cu_reduceMaxKernel<2>(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n);
template __global__ void _cu_reduceMaxKernel<1>(CorrelationPoint *g_idata, CorrelationPoint *g_odata, unsigned int n);

}	// namespace gpu
}	// namespace recognition
