#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <unistd.h>

#include <cuda_runtime.h>
#include <helper_cuda.h>
#include <helper_functions.h>

#include <log4cxx/logger.h>
#include <base/logging.h>

#include <cv.h>
#include <highgui.h>

#include <base/types.cuh>
#include <base/methods.cuh>
#include <gpu/cu_methods.cuh>

#include <gpu/fast_match_helper.cuh>
#include <gpu/fast_match.cuh>

namespace recognition {
namespace gpu {

using namespace cv;

const log4cxx::LoggerPtr BestTransformation::logger = log4cxx::LoggerPtr(log4cxx::Logger::getLogger("recognition.gpu.BestTransformation"));
const log4cxx::LoggerPtr BestTransformation::output = log4cxx::LoggerPtr(log4cxx::Logger::getLogger("output.recognition.gpu.BestTransformation"));

BestTransformationResult BestTransformation::FindBestTransformation() {
	BestTransformationResult res;

	double blur_sigma = 2;
	double blur_size = 4 * blur_sigma - 1;

	GaussianBlur(I1, I1, Size(blur_size, blur_size), blur_sigma, blur_sigma, BORDER_REFLECT);
	GaussianBlur(I2, I2, Size(blur_size, blur_size), blur_sigma, blur_sigma, BORDER_REFLECT);

	int numPoints = round(10 / (epsilon * epsilon));

	Mat xs = Mat(1, numPoints, CV_32SC1);
	Mat ys = Mat(1, numPoints, CV_32SC1);
	randu(xs, Scalar(1), Scalar(I1.cols + 1));
	randu(ys, Scalar(1), Scalar(I1.rows + 1));

	Mat configs;
	CreateListOfConfigs(configs);

	if (configs.rows > MAX_TRANSFORMATIONS) {
		LOG_ERROR(logger, "Too many transformations: %d", configs.rows);
		exit(1);
	}

	double deltaFact = 1.511;
	int level = 0;
	double newDelta = delta;
	double totTime = 0;

	int max_levels = 20;
	Mat bestDists(1, max_levels + 1, CV_64FC1);
	int nGoodConfigs = 0;

	double bestDist = -1;

	while (1) {
		++level;

		// Create affine matrixes
		double Configs2AffineTime = (double)getTickCount();
		printf("----- Configs2Affine, with %d configs -----\n", configs.rows);

		Mat affines(configs.rows, 6, CV_64FC1);
		Mat insiders(1, configs.rows, CV_32SC1);
		int n_in_bounds;

		Configs2Affine(configs, affines, insiders, n_in_bounds);

		cout << "n_in_bounds = " << n_in_bounds << endl;

		// TODO: check for memory leak
		Mat affines_filterd = Mat(n_in_bounds, 6, CV_64FC1);
		FilterByBounds(affines, affines_filterd, insiders);
		affines = affines_filterd;

		int oldGridSize = configs.rows;
		Mat configs_filterd = Mat(n_in_bounds, 6, CV_64FC1);
		FilterByBounds(configs, configs_filterd, insiders);
		configs = configs_filterd;

		Configs2AffineTime = ((double) getTickCount() - Configs2AffineTime) / getTickFrequency();

		//%% 2] evaluate all configurations
		double EvaluateConfigsTime = (double) getTickCount();

		Mat distances(1, n_in_bounds, CV_64FC1);
		EvaluateConfigsCuda(I1, I2, affines, xs, ys, photometricInvariance, distances);
		printf("----- Evaluate Configs, with %d configs -----\n", n_in_bounds);
		EvaluateConfigsTime = ((double) getTickCount() - EvaluateConfigsTime) / getTickFrequency();

		totTime += Configs2AffineTime + EvaluateConfigsTime;
		printf("Configs2AffineTime time: %f\n", Configs2AffineTime);
		printf("EvaluateConfigsTime time: %f\n", EvaluateConfigsTime);
		printf("Total time: %f\n", totTime);

		double* distances_ptr = distances.ptr<double>(0);
		bestDist = distances_ptr[0];
		int bestTransfIndex = 0;
		for (int i = 1; i < n_in_bounds; ++i) {
			if (distances_ptr[i] < bestDist) {
				bestDist = distances_ptr[i];
				bestTransfIndex = i;
			}
		}

		configs.row(bestTransfIndex).copyTo(res.bestConfig);
		CreateAffineTransformation(res.bestConfig, res.bestTransMat);
//			cout << res.bestConfig << endl;
//			cout << res.bestTransMat;

		bestDists.at<double>(0, level) = bestDist;

		fprintf(stderr, "$$$ bestDist = %.3f with index=%d\n", bestDist, bestTransfIndex);
		if ((bestDist < 0.005) || ((level > 2) && (bestDist < 0.015)) || (level >= max_levels) ||
				((level > 3) && (bestDist > mean<double>(bestDists, level - 3, level) * 0.97)) ||
				((level > 8) && (nGoodConfigs > max_levels)))
			break;

		//%% 5] choose the 'surviving' configs and delta for next round
		Mat goodConfigs;
		int tooHighPercentage;
		GetGoodConfigsByDistance(configs, bestDist, newDelta, distances, goodConfigs, tooHighPercentage);

		int numConfigs = goodConfigs.rows;

		//%% 7] expand 'surviving' configs for next round
		if ((tooHighPercentage && (bestDist > 0.05) && (level == 1) && (oldGridSize < 7500000)) ||
								((bestDist > 0.1) && (level == 1) && (oldGridSize < 5000000) ) ) {
			double fact = 0.9;
			fprintf(stderr, "##### RESTARTING!!! changing from delta: %.3f, to delta: %.3f\n", newDelta, newDelta * fact);
			newDelta = newDelta * fact;
			level = 0;
			steps.tx = fact * steps.tx;
			steps.ty = fact * steps.ty;
			steps.r = fact * steps.r;
			steps.s = fact * steps.s;
			CreateListOfConfigs(configs);
		} else {
			double prevDelta = newDelta;
			newDelta = newDelta / deltaFact;
			fprintf(stderr, "##### CONTINUING!!! prevDelta = %.3f,  newDelta = %.3f \n", prevDelta, newDelta);

			//%% expand the good configs
			Mat expandedConfigs;
			int npoints = 80;
			expansionType expandType = RANDOM_EXPANSION;
			switch (expandType) {
			case RANDOM_EXPANSION:
				ExpandConfigsRandom(goodConfigs, level, npoints, deltaFact, expandedConfigs);
				break;
			case FULL_EXPANSION:
				ExpandConfigsFull(goodConfigs, level, deltaFact, expandedConfigs);
				break;
			}

			printf("******************************** EXPANDED: %d\n", expandedConfigs.rows);

			goodConfigs.push_back(expandedConfigs);
			configs = goodConfigs;
		}

		printf("***\n");
		fprintf(stderr, "*** level %d:|goodConfigs| = %d, |expandedConfigs| = %d\n", level, numConfigs, configs.rows);
		printf("***\n");


		//%% 8] refresh random points
		// TODO:
		randu(xs, Scalar(1), Scalar(I1.cols + 1));
		randu(ys, Scalar(1), Scalar(I1.rows + 1));
	}

	res.sampledError = bestDist;

	return res;
}

void BestTransformation::EvaluateConfigsCuda(Mat& I1, Mat& I2, Mat& configs_, Mat& xs_, Mat& ys_, int photometricInvariance, Mat& distances_) {
	// parameters
	int h1, w1, h2, w2;
	int r1x, r1y, r2x, r2y;

	int numPoints, numConfigs;
	double a11, a12, a13, a21, a22, a23;

	// input variables
	double *img1, *img2;
	int *xs, *ys;

	// Should we use photometrics?
	int usePhotometrics = photometricInvariance;

	// helper variables
	int *xs_centered, *ys_centered;
	double *valsI1;
	int targetPoint_x, targetPoint_y;
	int targetInd;

	double* distances = distances_.ptr<double>(0);

	/* Find the dimensions of the data */
	h1 = I1.rows;
	w1 = I1.cols;
	h2 = I2.rows;
	w2 = I2.cols;
	numConfigs = configs_.rows;
	numPoints = xs_.cols;

	r1x = 0.5 * (w1 - 1);
	r1y = 0.5 * (h1 - 1);
	r2x = 0.5 * (w2 - 1);
	r2y = 0.5 * (h2 - 1);

	/* Create an mxArrays for temporary data */
	xs_centered = (int*) malloc(numPoints * sizeof(int));
	ys_centered = (int*) malloc(numPoints * sizeof(int));
	valsI1 = (double*) malloc(numPoints * sizeof(double));

	/* Store target pixel locations - x and y */
	double* xs_target = (double*) malloc(numPoints * sizeof(double));
	double* ys_target = (double*) malloc(numPoints * sizeof(double));

	/* Retrieve the input data */
	img1 = I1.ptr<double>(0);
	double* tmp_img2 = I2.ptr<double>(0);

	double* configs = configs_.ptr<double>(0);
	xs = xs_.ptr<int>(0);
	ys = ys_.ptr<int>(0);
	// img2 is of height 3*img2 (this padding is for not needing to check bounds)
	img2 = (double*) malloc(5 * h2 * w2 * sizeof(double));
	memset(img2, 2, 5 * h2 * w2 * sizeof(double));
	memcpy(img2 + 2 * h2 * w2, tmp_img2, h2 * w2 * sizeof(double));

	/*Centered pointes*/
	for (int i = 0 ; i < numPoints ; ++i) {
		xs_centered[i] = xs[i] - (r1x + 1);
		ys_centered[i] = ys[i] - (r1y + 1);
	}

	/*Precalculating source point indices into I1 (and the values themselves)*/
	for (int j = 0; j < numPoints ; j++) {
//	    	if (j < 10)
		valsI1[j] = img1[(ys[j] - 1) * w1 + (xs[j] - 1)]; // -1 is for c
	}

	int dimBlockSize = BLOCK_SIZE * BLOCK_SIZE;

	if (numConfigs > dimBlockSize) {

		// Load to device memory
		double *d_configs;
		checkCudaErrors(cudaMalloc((void **) &(d_configs), sizeof(double) * numConfigs * 6));
		checkCudaErrors(cudaMemcpy(d_configs, configs, sizeof(double) * numConfigs * 6, cudaMemcpyHostToDevice));

		int *d_xs_centered, *d_ys_centered;
		checkCudaErrors(cudaMalloc((void **) &(d_xs_centered), sizeof(int) * numPoints));
		checkCudaErrors(cudaMemcpy(d_xs_centered, xs_centered, sizeof(int) * numPoints, cudaMemcpyHostToDevice));

		checkCudaErrors(cudaMalloc((void **) &(d_ys_centered), sizeof(int) * numPoints));
		checkCudaErrors(cudaMemcpy(d_ys_centered, ys_centered, sizeof(int) * numPoints, cudaMemcpyHostToDevice));

		double *d_valsI1;
		checkCudaErrors(cudaMalloc((void **) &(d_valsI1), sizeof(double) * numPoints));
		checkCudaErrors(cudaMemcpy(d_valsI1, valsI1, sizeof(double) * numPoints, cudaMemcpyHostToDevice));

		double *d_img2;
		checkCudaErrors(cudaMalloc((void **) &(d_img2), sizeof(double) * 5 * h2 * w2));
		checkCudaErrors(cudaMemset(d_img2, 2, sizeof(double) * 5 * h2 * w2));
		checkCudaErrors(cudaMemcpy(d_img2 + 2 * h2 * w2, tmp_img2, sizeof(double) * h2 * w2, cudaMemcpyHostToDevice));

		double *d_distances;
		checkCudaErrors(cudaMalloc((void **) &(d_distances), sizeof(double) * numConfigs));
		// TODO: MAX_DOUBLE in case some indexes left unprocessed (which would be a bug!!!)
//		checkCudaErrors(cudaMemset(d_xs_target, xs_target, sizeof(int) * numPoints, cudaMemcpyHostToDevice));

		// Additional arguments
		_cu_findDistancesKernelArgs _args(numPoints, r2x, r2y, w2, h2, usePhotometrics);
		_cu_findDistancesKernelArgs *h_args = &_args;
		_cu_findDistancesKernelArgs *d_args;

		checkCudaErrors(cudaMalloc((void **) &d_args, sizeof(_cu_findDistancesKernelArgs)));
		checkCudaErrors(cudaMemcpy(d_args, h_args, sizeof(_cu_findDistancesKernelArgs), cudaMemcpyHostToDevice));

		// Find correlation
		dim3 dimBlock(BLOCK_SIZE * BLOCK_SIZE);
		dim3 dimGrid(numConfigs / dimBlock.x);
		_cu_findDistancesKernel<<<dimGrid, dimBlock>>>(d_configs, d_xs_centered, d_ys_centered, d_valsI1,
				d_img2, d_args, d_distances);
		getLastCudaError("Kernel execution failed");

		checkCudaErrors(cudaMemcpy(distances, d_distances, sizeof(double) * (numConfigs - numConfigs % dimBlock.x), cudaMemcpyDeviceToHost));

		cudaFree(d_args);
		cudaFree(d_distances);
		cudaFree(d_img2);
		cudaFree(d_valsI1);
		cudaFree(d_ys_centered);
		cudaFree(d_xs_centered);
		cudaFree(d_configs);
	}

	// Fill the rest distances
	if (numConfigs % dimBlockSize > 0) {
		for (int i = numConfigs - numConfigs % dimBlockSize ; i < numConfigs ; ++i) {
			configs = configs_.ptr<double>(i);
			a11 = configs[0];
			a12 = configs[1];
			a13 = configs[2];
			a21 = configs[3];
			a22 = configs[4];
			a23 = configs[5];

			double score = 0;

			double tmp1 = (r2x + 1) + a13 + 0.5;
			double tmp2 = (r2y + 1) + a23 + 0.5 + 2 * h2;
			double* ptrVals = valsI1;
			int* ptrXsc = xs_centered;
			int* ptrYsc = ys_centered;
			if (!usePhotometrics) {
				for (int j = 0; j < numPoints ; j++) {
					targetPoint_x = a11 * (*ptrXsc) + a12 * (*ptrYsc) + tmp1; // includes rounding
					targetPoint_y = a21 * (*ptrXsc) + a22 * (*ptrYsc) + tmp2; // includes rounding
					targetInd = (targetPoint_y - 1) * w2 + targetPoint_x - 1; // -1 is for c
					score += (fabs((*ptrVals) - img2[targetInd]));

					ptrVals++;
					ptrXsc++;
					ptrYsc++;
				}
			} else {
				double sumXiYi = 0; double sumXi = 0; double sumYi = 0;
				double sumXiSqrd = 0; double sumYiSqrd = 0; double Xi, Yi;

				for (int j = 0; j < numPoints ; j++) {
					targetPoint_x = a11*(*ptrXsc) + a12*(*ptrYsc) + tmp1; // includes rounding
					targetPoint_y = a21*(*ptrXsc) + a22*(*ptrYsc) + tmp2; // includes rounding
					targetInd = (targetPoint_y - 1) * w2 + (targetPoint_x - 1); // -1 for c
					Xi = (*ptrVals) ;
					Yi = img2[targetInd];
					xs_target[j] = Xi;
					ys_target[j] = Yi;

					sumXi += Xi;
					sumYi += Yi;
					sumXiSqrd += (Xi * Xi);
					sumYiSqrd += (Yi * Yi);
					ptrVals++;
					ptrXsc++;
					ptrYsc++;
				}
				// new score, based on normalizing mean and std of each of the signals
				double epsilon = 0.0000001;
				double sigX, sigY, meanX, meanY;
				sigX = sqrt((sumXiSqrd - (sumXi * sumXi) / numPoints) / numPoints) + epsilon;
				sigY = sqrt((sumYiSqrd - (sumYi * sumYi) / numPoints) / numPoints) + epsilon;
				meanX = sumXi / numPoints;
				meanY = sumYi / numPoints;
				double sigXoversigY = sigX / sigY;

				// Variable that stores a sum used repeatadly in the computation: -meanX+sigXoversigY*meanY
				double faster = -meanX + sigXoversigY * meanY;

				for (int j = 0; j < numPoints ; j++) {
					score += fabs(xs_target[j] - sigXoversigY * ys_target[j] + faster);
				}

			}
			distances[i] = score / numPoints;
		}
	}

	/* Free the allocated arrays */
	free(xs_centered);
	free(ys_centered);
	free(valsI1);
	free(xs_target);
	free(ys_target);
	free(img2);
}

__global__ void _cu_findDistancesKernel(double* d_configs, int* d_xs_centered, int* d_ys_centered, double* d_valsI1,
		double *d_img2, _cu_findDistancesKernelArgs *args, double *d_distances) {
	unsigned int tid = threadIdx.x;
	unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;

	int targetPoint_x, targetPoint_y, targetInd;
	double a11, a12, a13, a21, a22, a23;

	int numPoints = args->numPoints;
	int w2 = args->w2;
	int h2 = args->h2;
	int r2x = args->r2x;
	int r2y = args->r2y;
	int usePhotometrics = args->usePhotometrics;

	a11 = d_configs[6 * i];
	a12 = d_configs[6 * i + 1];
	a13 = d_configs[6 * i + 2];
	a21 = d_configs[6 * i + 3];
	a22 = d_configs[6 * i + 4];
	a23 = d_configs[6 * i + 5];

	double score = 0;

	double tmp1 = (r2x + 1) + a13 + 0.5;
	double tmp2 = (r2y + 1) + a23 + 0.5 + 2 * h2;
	double* ptrVals = d_valsI1;
	int* ptrXsc = d_xs_centered;
	int* ptrYsc = d_ys_centered;
	if (!usePhotometrics) {
		for (int j = 0; j < numPoints ; j++) {
			targetPoint_x = a11 * (*ptrXsc) + a12 * (*ptrYsc) + tmp1; // includes rounding
			targetPoint_y = a21 * (*ptrXsc) + a22 * (*ptrYsc) + tmp2; // includes rounding
			targetInd = (targetPoint_y - 1) * w2 + targetPoint_x - 1; // -1 is for c
			score += (fabs((*ptrVals) - d_img2[targetInd]));

			ptrVals++;
			ptrXsc++;
			ptrYsc++;
		}
	} else {
		/* Store target pixel locations - x and y */
		double sumXiYi = 0; double sumXi = 0; double sumYi = 0;
		double sumXiSqrd = 0; double sumYiSqrd = 0; double Xi, Yi;

		for (int j = 0; j < numPoints ; j++) {
			targetPoint_x = a11*(*ptrXsc) + a12*(*ptrYsc) + tmp1; // includes rounding
			targetPoint_y = a21*(*ptrXsc) + a22*(*ptrYsc) + tmp2; // includes rounding
			targetInd = (targetPoint_y - 1) * w2 + (targetPoint_x - 1); // -1 for c
			Xi = (*ptrVals) ;
			Yi = d_img2[targetInd];

			sumXi += Xi;
			sumYi += Yi;
			sumXiSqrd += (Xi * Xi);
			sumYiSqrd += (Yi * Yi);
			ptrVals++;
			ptrXsc++;
			ptrYsc++;
		}
		// new score, based on normalizing mean and std of each of the signals
		double epsilon = 0.0000001;
		double sigX, sigY, meanX, meanY;
		sigX = sqrt((sumXiSqrd - (sumXi * sumXi) / numPoints) / numPoints) + epsilon;
		sigY = sqrt((sumYiSqrd - (sumYi * sumYi) / numPoints) / numPoints) + epsilon;
		meanX = sumXi / numPoints;
		meanY = sumYi / numPoints;
		double sigXoversigY = sigX / sigY;

		// Variable that stores a sum used repeatadly in the computation: -meanX+sigXoversigY*meanY
		double faster = -meanX + sigXoversigY * meanY;

		// Attach to start of arrays again
		ptrVals = d_valsI1;
		ptrXsc = d_xs_centered;
		ptrYsc = d_ys_centered;

		// Calc again, since it's faster than dynamically allocate memory inside kernel to store temp result
		for (int j = 0; j < numPoints ; j++) {
			targetPoint_x = a11*(*ptrXsc) + a12*(*ptrYsc) + tmp1; // includes rounding
			targetPoint_y = a21*(*ptrXsc) + a22*(*ptrYsc) + tmp2; // includes rounding
			targetInd = (targetPoint_y - 1) * w2 + (targetPoint_x - 1); // -1 for c
			Xi = (*ptrVals) ;
			Yi = d_img2[targetInd];

			score += fabs(Xi - sigXoversigY * Yi + faster);

			ptrVals++;
			ptrXsc++;
			ptrYsc++;
		}
	}

	d_distances[i] = score / numPoints;
}

void BestTransformation::CreateAffineTransformation(cv::Mat& config, cv::Mat& transf) {
	double tx, ty, sx, sy, r1, r2;
	double* config_ptr = config.ptr<double>(0);

	tx =  config_ptr[0];
	ty =  config_ptr[1];
	r2 =  config_ptr[2];
	sx =  config_ptr[3];
	sy =  config_ptr[4];
	r1 =  config_ptr[5];

	//%% syms tx ty r2 sx sy r1
	//%% syms R1 T2 S A

	Mat R1 = (Mat_<double>(2, 2) << cos(r1), -sin(r1) , sin(r1), cos(r1));
	Mat R2 = (Mat_<double>(2, 2) << cos(r2), -sin(r2) , sin(r2), cos(r2));
	Mat S = (Mat_<double>(2, 2) << sx, 0 , 0, sy);
	Mat A = (Mat_<double>(3, 3) << 0, 0, tx , 0, 0, ty , 0, 0, 1);

//		A(1:2,1:2) = R1*S*R2;
	Mat A_ = R1 * S * R2;
	for (int y = 0; y < 2; ++y)
		for (int x = 0; x < 2; ++x)
			A.at<double>(y, x) = A_.at<double>(y, x);
//		cout << A << endl;

	transf = A;

//		% A =     [ sx*cos(r1)*cos(r2) - sy*sin(r1)*sin(r2), - sx*cos(r1)*sin(r2) - sy*cos(r2)*sin(r1), tx]
//		%         [ sx*cos(r2)*sin(r1) + sy*cos(r1)*sin(r2),   sy*cos(r1)*cos(r2) - sx*sin(r1)*sin(r2), ty]
//		%         [                                       0,                                         0,  1]
}

void BestTransformation::ExpandConfigsFull(Mat& configs, int level, double deltaFact, Mat& expandedConfigs) {
	double fact = pow(deltaFact, level);
	double halfstep_tx, halfstep_ty, halfstep_r, halfstep_s;
	halfstep_tx = steps.tx / fact;
	halfstep_ty = steps.ty / fact;
	halfstep_r = steps.r / fact;
	halfstep_s = steps.s / fact;

	int npoints = pow(3, 6);
	Mat additions = Mat::zeros(npoints, 6, CV_64FC1);
	double* additions_ptr;
	int i = 0;
	for (double tx = -halfstep_tx; tx <= halfstep_tx; tx += halfstep_tx)
		for (double ty = -halfstep_ty; ty <= halfstep_ty; ty += halfstep_ty)
			for (double r2 = -halfstep_r; r2 <= halfstep_r; r2 += halfstep_r)
				for (double sx = -halfstep_s; sx <= halfstep_s; sx += halfstep_s)
					for (double sy = -halfstep_s; sy <= halfstep_s; sy += halfstep_s)
						for (double r1 = -halfstep_r; r1 <= halfstep_r; r1 += halfstep_r) {
							additions_ptr = additions.ptr<double>(i);
							additions_ptr[0] = tx;
							additions_ptr[1] = ty;
							additions_ptr[2] = r2;
							additions_ptr[3] = sx;
							additions_ptr[4] = sy;
							additions_ptr[5] = r1;
							i = i + 1;
						}

	int numConfigs = configs.rows;
//		expanded= repmat(configs,[npoints,1]);
	Mat expanded = repeat(configs, npoints, 1);

//		expandedConfigs = expanded + repmat(additions,[numConfigs,1]);
	additions = repeat(additions, numConfigs, 1);
	add(expanded, additions, expandedConfigs);
}


void BestTransformation::ExpandConfigsRandom(Mat& configs, int level, int npoints, double deltaFact, Mat& expandedConfigs) {
	double fact = pow(deltaFact, level);
	double halfstep_tx, halfstep_ty, halfstep_r, halfstep_s;
	halfstep_tx = steps.tx / fact;
	halfstep_ty = steps.ty / fact;
	halfstep_r = steps.r / fact;
	halfstep_s = steps.s / fact;

	int numConfigs = configs.rows;

//		randvec = floor(3 * rand(npoints * numConfigs, 6) - 1); //%% random vectors in {-1,0,1}^6;
	Mat randvec_ = Mat(npoints * numConfigs, 6, CV_32SC1);
	randu(randvec_, Scalar(-1), Scalar(2));
	Mat randvec(randvec_.size(), CV_64FC1);
	randvec_.convertTo(randvec, CV_64FC1);

//		expanded= repmat(configs,[npoints,1]);
	Mat expanded = repeat(configs, npoints, 1);

//		ranges = [halfstep_tx,halfstep_ty,halfstep_r,halfstep_s,halfstep_s,halfstep_r];
	Mat ranges = (Mat_<double>(1, 6) << halfstep_tx, halfstep_ty, halfstep_r, halfstep_s, halfstep_s, halfstep_r);
//		Mat ranges = (Mat_<double>(1, 6) << 0.5, 0.5, 0.5, 0.5, 0.5, 0.5);

//		expandedConfigs = expanded + randvec.*repmat(ranges,[npoints*numConfigs,1]);
	ranges = repeat(ranges, npoints * numConfigs, 1);
	multiply(randvec, ranges, ranges);
	add(expanded, ranges, expandedConfigs);
}

//	function [thresh,goodConfigs,tooHighPercentage,extremelyHighPercentage] = GetGoodConfigsByDistance(configs,bestDist,newDelta,distances,bestGridVec)
void BestTransformation::GetGoodConfigsByDistance(Mat& configs, double bestDist, double newDelta, Mat& distances,
			Mat& goodConfigs, int& tooHighPercentage) {
	double thresh = bestDist + GetThreshPerDelta(newDelta);

	double* distances_ptr = distances.ptr<double>(0);
	std::vector<int> indexes;
	int nGoodConfigs = 0;
	for (int i = 0; i < distances.cols; ++i) {
		if (distances_ptr[i] <= thresh) {
			indexes.push_back(i);
			++nGoodConfigs;
		}
	}

	goodConfigs = Mat(nGoodConfigs, 6, CV_64FC1);

	for (int i = 0; i < nGoodConfigs; ++i) {
		double* goodConfigs_ptr = goodConfigs.ptr<double>(i);
		double* configs_ptr = configs.ptr<double>(indexes[i]);
		for (int j = 0; j < 6; ++j) {
			goodConfigs_ptr[j] = configs_ptr[j];
		}
	}

	int numConfigs = nGoodConfigs;
	double percentage = (double) numConfigs / configs.rows;

	fprintf(stderr, "percentage = %f\n", percentage);

	// TODO: come up with faster way to reduce (like binary search from bestDist to bestDist + GetThreshPerDelta(newDelta))
	while (numConfigs > 27000) {
		thresh = thresh * 0.99;

		numConfigs = 0;
		for (int i = 0; i < distances.cols; ++i)
			if (distances_ptr[i] <= thresh)
				++nGoodConfigs;
	}

	if (numConfigs == 0) {
		thresh = bestDist;

		numConfigs = 0;
		for (int i = 0; i < distances.cols; ++i)
			if (distances_ptr[i] <= thresh)
				++nGoodConfigs;

		if (nGoodConfigs > 10000) {
			std::vector<int> indexes;
			int numConfigs = 0;
			//%% all with the same error exactly - probably equivalent
			for (int i = 0; i < distances.cols; ++i) {
				if (numConfigs >= 100)
					break;

				if (distances_ptr[i] <= thresh) {
					indexes.push_back(i);
					++numConfigs;
				}
			}

			goodConfigs = Mat(numConfigs, 6, CV_64FC1);

			for (int i = 0; i < numConfigs; ++i) {
				double* goodConfigs_ptr = goodConfigs.ptr<double>(i);
				double* configs_ptr = configs.ptr<double>(indexes[i]);
				for (int j = 0; j < 6; ++j) {
					goodConfigs_ptr[j] = configs_ptr[j];
				}
			}
		}
	}

	tooHighPercentage = percentage > 0.022;
}


double BestTransformation::GetThreshPerDelta(double delta) {
	//%% based on experimentally drawn % p = 0.1341    0.0278
	double p1 = 0.1341;
	double p2 = 0.0278;
	double safety = 0.02;
	return p1 * delta + p2 - safety; //%%+0.01;
}


void BestTransformation::CreateListOfConfigs(Mat& configs) {
	int ntx_steps, nty_steps, nr_steps, ns_steps;
	int gridSize;

	ntx_steps = (bounds_max.tx - bounds_min.tx) / steps.tx + 1;
	nty_steps = (bounds_max.ty - bounds_min.ty) / steps.ty + 1;
	nr_steps = (bounds_max.r - bounds_min.r) / steps.r + 1;

	if (steps.s == 0)
		ns_steps = 1;
	else
		ns_steps = (bounds_max.s - bounds_min.s) / steps.s + 1;


	if (bounds_max.tx - bounds_min.tx - steps.tx * (ntx_steps - 1) > 0.5 * steps.tx)
		++ntx_steps;

	if (bounds_max.ty - bounds_min.ty - steps.ty * (nty_steps - 1) > 0.5 * steps.ty)
		++nty_steps;

	if (bounds_max.r - bounds_min.r - steps.r * (nr_steps - 1) > steps.r)
		++nr_steps;

	if (bounds_max.s - bounds_min.s - steps.s * (ns_steps - 1) > 0.5 * steps.s)
		++ns_steps;

	Mat tx_steps = Mat(1, ntx_steps, CV_64FC1);
	Mat ty_steps = Mat(1, nty_steps, CV_64FC1);
	Mat r_steps = Mat(1, nr_steps, CV_64FC1);
	Mat s_steps = Mat(1, ns_steps, CV_64FC1);

	generateRangeWithStep<double>(tx_steps, bounds_min.tx, steps.tx);
	generateRangeWithStep<double>(ty_steps, bounds_min.ty, steps.ty);
	generateRangeWithStep<double>(r_steps, bounds_min.r, steps.r);
	generateRangeWithStep<double>(s_steps, bounds_min.s, steps.s);

	int NR2_steps = nr_steps;

	if (bounds_max.r - bounds_min.r == 2 * M_PI) {
		NR2_steps = 0;
		MatIterator_<double> it, end;
		double max_bound = -M_PI / 2 + steps.r / 2;
		for (it = r_steps.begin<double>(), end = r_steps.end<double>(); it != end; ++it) {
			if (*it < max_bound) {
				NR2_steps++;
			}
		}
	}

	gridSize = ntx_steps * nty_steps * (ns_steps * ns_steps) * (nr_steps * NR2_steps);
	configs = Mat::zeros(gridSize, 6, CV_64FC1);

	CreateList(ntx_steps, nty_steps, nr_steps, NR2_steps, ns_steps, tx_steps, ty_steps, r_steps, s_steps, configs);
//		for (int i = 1; i < 13; i++)
//			cout << configs.row(gridSize - i) << endl;
//		cout << gridSize << endl;;
}

void BestTransformation::CreateList(int ntx_steps, int nty_steps, int nr_steps, int NR2_steps, int ns_steps,
			Mat& tx_steps, Mat& ty_steps, Mat& r_steps, Mat& s_steps, Mat& configs) {
	int row = 0;
	double* tx = tx_steps.ptr<double>(0);
	double* ty = ty_steps.ptr<double>(0);
	double* r = r_steps.ptr<double>(0);
	double* s = s_steps.ptr<double>(0);
	for (int tx_idx = 0; tx_idx < ntx_steps; ++tx_idx) {
		for (int ty_idx = 0; ty_idx < nty_steps; ++ty_idx) {
			for (int r1_idx = 0; r1_idx < nr_steps; ++r1_idx) {
				for (int r2_idx = 0; r2_idx < NR2_steps; ++r2_idx) {
					for (int sx_idx = 0; sx_idx < ns_steps; ++sx_idx) {
						for (int sy_idx = 0; sy_idx < ns_steps; ++sy_idx) {
							double* config = configs.ptr<double>(row);
							config[0] = tx[tx_idx];
							config[1] = ty[ty_idx];
							config[2] = r[r2_idx];
							config[3] = s[sx_idx];
							config[4] = s[sy_idx];
							config[5] = r[r1_idx];

							++row;
						}
					}
				}
			}
		}
	}
}

void BestTransformation::Configs2Affine(cv::Mat& configs, Mat& affines, Mat& insiders, int &n_in_bounds) {
	int h1, w1, h2, w2;
	int r1x, r1y, r2x, r2y;

	h1 = I1.rows;
	w1 = I1.cols;
	h2 = I2.rows;
	w2 = I2.cols;

	r1x = 0.5 * (w1 - 1);
	r1y = 0.5 * (h1 - 1);
	r2x = 0.5 * (w2 - 1);
	r2y = 0.5 * (h2 - 1);

	n_in_bounds = 0;
	int* insiders_ptr = insiders.ptr<int>(0);
	for (int row = 0; row < configs.rows ; ++row) {
		double* config = configs.ptr<double>(row);
		double tx, ty, r1, r2, sx, sy;
		tx = config[0];
		ty = config[1];
		r2 = config[2];
		sx = config[3];
		sy = config[4];
		r1 = config[5];

		//syms tx ty r2 sx sy r1;
		// A =     [ sx*cos(r1)*cos(r2) - sy*sin(r1)*sin(r2), - sx*cos(r1)*sin(r2) - sy*cos(r2)*sin(r1), tx]
		//         [ sx*cos(r2)*sin(r1) + sy*cos(r1)*sin(r2),   sy*cos(r1)*cos(r2) - sx*sin(r1)*sin(r2), ty]
		//         [                                       0,                                         0,  1]
		//        matrixConfigs(i,:) = [A(1,1),A(1,2),A(1,3),A(2,1),A(2,2),A(2,3)];

		double* affine = affines.ptr<double>(row);

		double a11, a12, a13, a21, a22, a23;

		affine[0] = a11 = sx * cos(r1) * cos(r2) - sy * sin(r1) * sin(r2);
		affine[1] = a12 = - sx * cos(r1) * sin(r2) - sy * cos(r2) * sin(r1);
		affine[2] = a13 = tx;
		affine[3] = a21 = sx * cos(r2) * sin(r1) + sy * cos(r1) * sin(r2);
		affine[4] = a22 = sy * cos(r1) * cos(r2) - sx * sin(r1) * sin(r2);
		affine[5] = a23 = ty;

		//mexPrintf("%.3f, %.3f, %.3f, %.3f, %.3f, %.3f\n",a11,a12,a13,a21,a22,a23);

		double c1x = a11 * (1 - (r1x + 1)) + a12 * (1 - (r1y + 1)) + (r2x + 1) + a13;
		double c1y = a21 * (1 - (r1x + 1)) + a22 * (1 - (r1y + 1)) + (r2y + 1) + a23;
		double c2x = a11 * (w1 - (r1x + 1)) + a12 * (1 - (r1y + 1)) + (r2x + 1) + a13;
		double c2y = a21 * (w1 - (r1x + 1)) + a22 * (1 - (r1y + 1)) + (r2y + 1) + a23;
		double c3x = a11 * (w1 - (r1x + 1)) + a12 * (h1 - (r1y + 1)) + (r2x + 1) + a13;
		double c3y = a21 * (w1 - (r1x + 1)) + a22 * (h1 - (r1y + 1)) + (r2y + 1) + a23;
		double c4x = a11 * (1 - (r1x + 1)) + a12 * (h1 - (r1y + 1)) + (r2x + 1) + a13;
		double c4y = a21 * (1 - (r1x + 1)) + a22 * (h1 - (r1y + 1)) + (r2y + 1) + a23;

		// allow to exceed boundary by at most q
		int q = 10;
		int k = (c1x > -q) && (c1x < w2 + q) && (c1y > -q) && (c1y < h2 + q) &&
				(c2x > -q) && (c2x < w2 + q) && (c2y > -q) && (c2y < h2 + q) &&
				(c3x > -q) && (c3x < w2 + q) && (c3y > -q) && (c3y < h2 - q) &&	// h2 - q --- BUG? should be h2 + q (bottom vertical boundary)
				(c4x > -q) && (c4x < w2 + q) && (c4y > -q) && (c4y < h2 + q);

		n_in_bounds += k;
		insiders_ptr[row] = k;
	}
}

void BestTransformation::FilterByBounds(cv::Mat& src, cv::Mat& dst, cv::Mat& filter) {
	int dst_row = 0;
	int* filter_ptr = filter.ptr<int>(0);

	for (int row = 0; row < src.rows; ++row) {
		if (filter_ptr[row] == 1) {
			double* src_ptr = src.ptr<double>(row);
			double* dst_ptr = dst.ptr<double>(dst_row);

			for (int i = 0; i < 6; i++)
				dst_ptr[i] = src_ptr[i];

			++dst_row;
		}
	}
}

//	void BestTransformation::EvaluateConfigs2(
//			int n_transformations,
//			int n_points,
//			double* pattern,		// n_points
//			double* image,			// 5 * h2 * w2
//			double* cover,			// 6 * n_transformations (column wise)
//			int w2,
//			double c,
//			int* x,					// n_points
//			int* y,					// n_points
//			double* correlations	// 6 * n_transformations
//			) {
//
//	    for (int i = 0 ; i < n_transformations ; ++i) {
//	    	double a11, a12, a13, a21, a22, a23;
//	    	double correlation = 0;
//
//			a11 = cover[6 * i];
//			a12 = cover[6 * i + 1];
//			a13 = cover[6 * i + 2];
//			a21 = cover[6 * i + 3];
//			a22 = cover[6 * i + 4];
//			a23 = cover[6 * i + 5];
//
//			for (int j = 0; j < n_points ; j++) {
//				int x_tmp, y_tmp, idx;
//
//				x_tmp = a11 * (*x) + a12 * (*y);
//				y_tmp = a21 * (*x) + a22 * (*y);
//				idx = y_tmp * w2 + x_tmp + a13 + a23 * w2 + c;
//				correlation += fabs((*pattern) - image[idx]);
//
//				++pattern;
//				++x;
//				++y;
//			}
//
//			correlations[i] = correlation / n_points;
//		}
//	}

void BestTransformation::EvaluateConfigs(Mat& I1, Mat& I2, Mat& configs_, Mat& xs_, Mat& ys_, int photometricInvariance, Mat& distances_) {
	// parameters
	int h1, w1, h2, w2;
	int r1x, r1y, r2x, r2y;

	int numPoints, numConfigs;
	double a11, a12, a13, a21, a22, a23;

	// input variables
	double *img1, *img2;
	int *xs, *ys;

	// Should we use photometrics?
	int usePhotometrics = photometricInvariance;

	// helper variables
	int *xs_centered, *ys_centered;
	double *valsI1;
	int targetPoint_x, targetPoint_y;
	int targetInd;

	double sigX;
	double sigY;
	double meanX;
	double meanY;

	double* distances = distances_.ptr<double>(0);

	// inputs:     dimensions
	// ------	    -----------
	// image I1     h1 x w1
	// image I2     h2 x w2
	// configs      numConfigs x 6
	// xs           1 x numPoints
	// ys1			1 x numPoints

	// outputs:     dimensions
	// -------	    -----------
	// distances    1 x numConfigs


	/* Find the dimensions of the data */
	h1 = I1.rows;
	w1 = I1.cols;
	h2 = I2.rows;
	w2 = I2.cols;
//	    h1 = I1.cols;
//	    w1 = I1.rows;
//	    h2 = I2.cols;
//	    w2 = I2.rows;
	numConfigs = configs_.rows;
	numPoints = xs_.cols;

	r1x = 0.5 * (w1 - 1);
	r1y = 0.5 * (h1 - 1);
	r2x = 0.5 * (w2 - 1);
	r2y = 0.5 * (h2 - 1);

	/* Create an mxArrays for temporary data */
	xs_centered = (int*) malloc(numPoints * sizeof(int));
	ys_centered = (int*) malloc(numPoints * sizeof(int));
	valsI1 = (double*) malloc(numPoints * sizeof(double));

	/* Store target pixel locations - x and y */
	double* xs_target = (double*) malloc(numPoints * sizeof(double));
	double* ys_target = (double*) malloc(numPoints * sizeof(double));

	/* Retrieve the input data */
	img1 = I1.ptr<double>(0);
	double* tmp_img2 = I2.ptr<double>(0);

	double* configs = configs_.ptr<double>(0);
	xs = xs_.ptr<int>(0);
	ys = ys_.ptr<int>(0);
	// img2 is of height 3*img2 (this padding is for not needing to check bounds)
	img2 = (double*) malloc(5 * h2 * w2 * sizeof(double));
	memset(img2, 2, 5 * h2 * w2 * sizeof(double));
	memcpy(img2 + 2 * h2 * w2, tmp_img2, h2 * w2 * sizeof(double));

	/*Centered pointes*/
	for (int i = 0 ; i < numPoints ; ++i) {
		xs_centered[i] = xs[i] - (r1x + 1);
		ys_centered[i] = ys[i] - (r1y + 1);
	}

	/*Precalculating source point indices into I1 (and the values themselves)*/
	for (int j = 0; j < numPoints ; j++) {
		valsI1[j] = img1[(ys[j] - 1) * w1 + (xs[j] - 1)]; // -1 is for c
	}

	for (int i = 0 ; i < numConfigs ; ++i) {
		configs = configs_.ptr<double>(i);
		a11 = configs[0];
		a12 = configs[1];
		a13 = configs[2];
		a21 = configs[3];
		a22 = configs[4];
		a23 = configs[5];

		double score = 0;

		double tmp1 = (r2x + 1) + a13 + 0.5;
		double tmp2 = (r2y + 1) + a23 + 0.5 + 2 * h2;
		double* ptrVals = valsI1;
		int* ptrXsc = xs_centered;
		int* ptrYsc = ys_centered;
		if (!usePhotometrics) {
			for (int j = 0; j < numPoints ; j++) {
				targetPoint_x = a11 * (*ptrXsc) + a12 * (*ptrYsc) + tmp1; // includes rounding
				targetPoint_y = a21 * (*ptrXsc) + a22 * (*ptrYsc) + tmp2; // includes rounding
				targetInd = (targetPoint_y - 1) * w2 + targetPoint_x - 1; // -1 is for c
				score += (fabs((*ptrVals) - img2[targetInd]));

				ptrVals++;
				ptrXsc++;
				ptrYsc++;
			}
		} else {
			double sumXiYi = 0; double sumXi = 0; double sumYi = 0;
			double sumXiSqrd = 0; double sumYiSqrd = 0; double Xi, Yi;

			for (int j = 0; j < numPoints ; j++) {
				targetPoint_x = a11*(*ptrXsc) + a12*(*ptrYsc) + tmp1; // includes rounding
				targetPoint_y = a21*(*ptrXsc) + a22*(*ptrYsc) + tmp2; // includes rounding
				targetInd = (targetPoint_y - 1) * w2 + (targetPoint_x - 1); // -1 for c
				Xi = (*ptrVals) ;
				Yi = img2[targetInd];
				xs_target[j] = Xi;
				ys_target[j] = Yi;

				sumXi += Xi;
				sumYi += Yi;
				sumXiSqrd += (Xi * Xi);
				sumYiSqrd += (Yi * Yi);
				ptrVals++;
				ptrXsc++;
				ptrYsc++;
			}
			// old score based on correlation
//                         score = numPoints*sumXiYi-sumXi*sumYi;
//                         score /= (0.00001 + sqrt(numPoints*sumXiSqrd - sumXi*sumXi) * sqrt(numPoints*sumYiSqrd - sumYi*sumYi) );
//                         score  = (1-score)/2;
// 			score *= numPoints;
			// new score, based on normalizing mean and std of each of the signals
			double epsilon = 0.0000001;
			sigX = sqrt((sumXiSqrd - (sumXi * sumXi) / numPoints) / numPoints) + epsilon;
			sigY = sqrt((sumYiSqrd - (sumYi * sumYi) / numPoints) / numPoints) + epsilon;
			meanX = sumXi / numPoints;
			meanY = sumYi / numPoints;
			double sigXoversigY = sigX / sigY;

			// Variable that stores a sum used repeatadly in the computation: -meanX+sigXoversigY*meanY
			double faster = -meanX + sigXoversigY * meanY;

			for (int j = 0; j < numPoints ; j++) {
//                     			  score += fabs( (xs_target[j]-meanX)/sigX - (ys_target[j]-meanY)/sigY);
//                                score += fabs( (xs_target[j]-meanX) - sigX*(ys_target[j]-meanY)/sigY);
//                                score += fabs( (xs_target[j]-meanX) - sigXoversigY*(ys_target[j]-meanY));
				score += fabs(xs_target[j] - sigXoversigY * ys_target[j] + faster);
			}

		}
		distances[i] = score / numPoints;

	}

	/* Free the allocated arrays */
	free(xs_centered);
	free(ys_centered);
	free(valsI1);
	free(xs_target);
	free(ys_target);
	free(img2);
}

}	// namespace gpu
}	// namespace recognition
