#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <unistd.h>

#include <cuda_runtime.h>
#include <helper_cuda.h>
#include <helper_functions.h>

#include <log4cxx/logger.h>
#include <base/logging.h>

#include <base/types.cuh>
#include <base/methods.cuh>
#include <gpu/cu_methods.cuh>
#include <gpu/gpu_correlation2.cuh>

namespace recognition {
namespace gpu {

const log4cxx::LoggerPtr GPUCorrelation2::logger = log4cxx::LoggerPtr(log4cxx::Logger::getLogger("recognition.gpu.GPUCorrelation2"));
const log4cxx::LoggerPtr GPUCorrelation2::output = log4cxx::LoggerPtr(log4cxx::Logger::getLogger("output.recognition.gpu.GPUCorrelation2"));

// Image correlation kernel:
// A - camera, B - pattern, C - each elements of matrix is a correlation of A and B
__global__ void _cu_findCorrelationKernel(Matrix2f *A, SparseImage *B, CorrelationPoint *C, _cu_findCorrelationKernelArgs *args) {
	int blockRow = blockIdx.y;
	int blockCol = blockIdx.x;

	int row = threadIdx.y;
	int col = threadIdx.x;

	int x = blockCol * BLOCK_SIZE + col;
	int y = blockRow * BLOCK_SIZE + row;

	// Each thread computes one element of C by accumulating results into corr
	img_t corr = 0;

	img_t avr = 0;
	img_t norm = 0;

	PointList start = B->list->ofSelf();
	PointList *list = &start;

	// Loop over all the sub-matrices of A and B that are required to
	// compute Csub. Find correlation between each pair of sub-matrices
	// and accumulate the results
	int cornerIdx = y * A->stride + x;
	while ((list = list->next) != NULL) {
		int val = A->elements[cornerIdx + list->y * A->stride + list->x];
		avr += val;
		norm += val * val;
		corr += B->elements[list->y * B->stride + list->x] * val;
	}

	avr = avr / B->n_elements;
	norm = norm / B->n_elements;
	norm -= avr * avr;
	norm = norm != 0 ? sqrt(norm) : 1;

	corr = (corr - avr * args->pattern_sum) / (norm * B->n_elements);

//	if (x == 356 && y == 137) {
//		printf("\n>>>>>>>(%d, %d) : Cvalue=%f, A[x,y]=%f", x, y, Cvalue, A->elements[y * A->stride + x]);
//	}

	atomicAdd(&(args->debug), 1.0F);

	// Write Csub to device memory. Each thread writes one element
	CorrelationPoint res;
	res.y = y;
	res.x = x;
	res.corr = corr;
	C[y * (A->stride - B->stride) + x] = res;

//	__syncthreads();
}

// DEPRECATED
__global__ void _cu_findMaximumKernel(CorrelationPoint *g_idata, CorrelationPoint *g_odata) {
	__shared__ CorrelationPoint sdata[N_THREADS_FOR_REDUCE];

	// Each thread loads one element from global to shared mem
	unsigned int tid = threadIdx.x;
	unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;

	sdata[tid] = g_idata[i];
	__syncthreads();

	// Do reduction in shared mem
	for (unsigned int s = blockDim.x / 2; s > 0; s >>= 1) {
		if (tid < s) {
			sdata[tid] = sdata[tid + s].corr >= sdata[tid].corr ? sdata[tid + s] : sdata[tid];
		}

		__syncthreads();
	}

	// Write result for this block to global mem
	if (tid == 0) g_odata[blockIdx.x] = sdata[0];
}

// TODO: implement border case for x = 386 || y = 386 (x, y start from 0 and end 511)
// Image dimensions are assumed to be multiples of BLOCK_SIZE
CorrelationResult GPUCorrelation2::FindCorrelation(Matrix2f *d_camera, SparseImage pattern, img_t pattern_sum, int &x_out, int &y_out) {
	// Load pattern to device memory
	SparseImage *h_pattern = &pattern;
	SparseImage *d_pattern;
	PointList *d_pattern_list;
	img_t *d_pattern_elements;

	// See http://stackoverflow.com/questions/9309195/copying-a-struct-containing-pointers-to-cuda-device
	// for more information about copying structures from host to device memory back and forth.

	// Allocate struct
	checkCudaErrors(cudaMalloc((void **) &d_pattern, sizeof(SparseImage)));

	// Allocate and copy list
	checkCudaErrors(cudaMalloc((void **) &d_pattern_list, sizeof(PointList)));

	PPointList *h_ppList = (PPointList*) malloc(sizeof(PPointList)); // Tracks any allocation in device memory
	PPointList *ppList = h_ppList;
	PointList *current = d_pattern_list;
	PointList startElement = h_pattern->list->ofSelf();
	PointList *list = &startElement;
	while ((list = list->next) != NULL) {
		PointList *next = NULL;
		ppList->list = current;

		if (list->next != NULL) {
			checkCudaErrors(cudaMalloc((void **) &(next), sizeof(PointList)));

			ppList->next = (PPointList*) malloc(sizeof(PPointList));
		} else {
			ppList->next = NULL;
		}

		PointList data(list->x, list->y, next);

		checkCudaErrors(cudaMemcpy(current, &data, sizeof(PointList), cudaMemcpyHostToDevice));

		current = next;
		ppList = ppList->next;
	}

	// Allocate and copy elements
	checkCudaErrors(cudaMalloc((void **) &(d_pattern_elements), sizeof(img_t) * h_pattern->stride * h_pattern->height));
	checkCudaErrors(cudaMemcpy(d_pattern_elements, h_pattern->elements, sizeof(img_t) * h_pattern->stride * h_pattern->height, cudaMemcpyHostToDevice));

	// Copy struct
	checkCudaErrors(cudaMemcpy(d_pattern, h_pattern, sizeof(SparseImage), cudaMemcpyHostToDevice));

	// Copy pointers
	// Note, that dereferencing of device pointer on host works only with cudaMemcpy()
	checkCudaErrors(cudaMemcpy(&(d_pattern->list), &d_pattern_list, sizeof(PointList*), cudaMemcpyHostToDevice));
	checkCudaErrors(cudaMemcpy(&(d_pattern->elements), &d_pattern_elements, sizeof(img_t*), cudaMemcpyHostToDevice));

	// Allocate correlation matrix in device memory
	CorrelationPoint *g_idata;

	checkCudaErrors(cudaMalloc((void **) &g_idata,
			sizeof(CorrelationPoint) * (camera.height - h_pattern->height) * (camera.stride - h_pattern->stride)));

	// Additional arguments
	_cu_findCorrelationKernelArgs _args(pattern_sum);
	_cu_findCorrelationKernelArgs *h_args = &_args;
	_cu_findCorrelationKernelArgs *d_args;

	checkCudaErrors(cudaMalloc((void **) &d_args, sizeof(_cu_findCorrelationKernelArgs)));
	checkCudaErrors(cudaMemcpy(d_args, h_args, sizeof(_cu_findCorrelationKernelArgs), cudaMemcpyHostToDevice));

	// Allocate CUDA events that we'll use for timing
	cudaEvent_t start;
	checkCudaErrors(cudaEventCreate(&start));

	cudaEvent_t stop;
	checkCudaErrors(cudaEventCreate(&stop));

	// Record the start event
	checkCudaErrors(cudaEventRecord(start, NULL));

	// Find correlation
	dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
	dim3 dimGrid((camera.stride - h_pattern->stride) / dimBlock.x, (camera.height - h_pattern->height) / dimBlock.y);
	_cu_findCorrelationKernel<<<dimGrid, dimBlock>>>(d_camera, d_pattern, g_idata, d_args);
	getLastCudaError("Kernel execution failed");

	/////////////////// Find maximum correlation in result matrix ////////////////////////
	// Compute max correlation
	int n_threads = N_THREADS_FOR_REDUCE;
	int n_elements = (camera.height - h_pattern->height) * (camera.stride - h_pattern->stride);
	// reduces the number of blocks, therefore burdens threads with more cycle iterations
	int multiplier = 16;
	int n_blocks = n_elements / (n_threads * multiplier);
	int size = n_blocks;

	// Allocate reduce output array
	CorrelationPoint *g_odata;
	checkCudaErrors(cudaMalloc((void **) &g_odata, sizeof(CorrelationPoint) * size));

	// Look for maximum correlation
	{
		dim3 dimBlock(n_threads);
		dim3 dimGrid(n_blocks);
		unsigned int smemSize = dimBlock.x;	// should work with dynamic shared mem allocation

		switch (dimBlock.x) {
		case 1024:
			_cu_reduceMaxKernel<1024><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 512:
			_cu_reduceMaxKernel< 512><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 256:
			_cu_reduceMaxKernel< 256><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 128:
			_cu_reduceMaxKernel< 128><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 64:
			_cu_reduceMaxKernel<  64><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 32:
			_cu_reduceMaxKernel<  32><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 16:
			_cu_reduceMaxKernel<  16><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 8:
			_cu_reduceMaxKernel<   8><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 4:
			_cu_reduceMaxKernel<   4><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 2:
			_cu_reduceMaxKernel<   2><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 1:
			_cu_reduceMaxKernel<   1><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		}
	}

//	_cu_findMaximum<<<n_blocks, n_threads>>>(g_idata, g_odata);
	CorrelationPoint *data = (CorrelationPoint*) malloc(sizeof(CorrelationPoint) * size);

	checkCudaErrors(cudaMemcpy(data, g_odata, sizeof(CorrelationPoint) * size, cudaMemcpyDeviceToHost));

	CorrelationPoint cPoint;
	cPoint.x = cPoint.y = -1;
	cPoint.corr = 0;

	for (int i = 0; i < size; ++i) {
		CorrelationPoint cur_corr = data[i];
		if (cPoint.corr < cur_corr.corr) {
			cPoint = cur_corr;
		}
	}

	img_t corr = cPoint.corr;
	int y_res = y_out = cPoint.y;
	int x_res = x_out = cPoint.x;

	free(data);
	//////////////////////////////////////////////////////////////////////////////////////

	 // Record the stop event
	checkCudaErrors(cudaEventRecord(stop, NULL));

	// Wait for the stop event to complete
	checkCudaErrors(cudaEventSynchronize(stop));

	float msecTotal = 0.0f;
	checkCudaErrors(cudaEventElapsedTime(&msecTotal, start, stop));
	float sec = msecTotal / 1000.0F;

	_cu_findCorrelationKernelArgs _debug(0);
	_cu_findCorrelationKernelArgs *debug = &_debug;
	checkCudaErrors(cudaMemcpy(debug, d_args, sizeof(_cu_findCorrelationKernelArgs), cudaMemcpyDeviceToHost));

//	printf("\nDebug: %f", debug->debug);

	CorrelationResult res;
	res.time = sec;
	res.deviation = sqrt(pow(x_res - x_target, 2) + pow(y_res - y_target, 2));

	LOG_INFO(logger, __FMT_SIMULATION_ITERATION_RESULT, iteration_, skip_, sec, corr, x_res, y_res);

	// Cleanup memory
	checkCudaErrors(cudaFree(g_odata));
	checkCudaErrors(cudaFree(g_idata));
	checkCudaErrors(cudaFree(d_args));
	h_ppList->_cu_freeList();
	checkCudaErrors(cudaFree(d_pattern_elements));
	checkCudaErrors(cudaFree(d_pattern));
	return res;
}

// Estimating average time and deviation with fixed probability
void GPUCorrelation2::average(unsigned samples, float from, float to) {
	if (samples <= 0) {
		LOG_ERROR(logger, "average takes positive number of samples");
		exit(EXIT_FAILURE);
	}

	if (from > to || from < 0.0F || to >= 1.0F) {
		LOG_ERROR(logger, "average takes from and to in range [0, 1) and from should be not greater than to");
		exit(EXIT_FAILURE);
	}

	int from_int = (int) (from * 10);
	int to_int = (int) (to * 10);

	// Load camera to device memory
	Matrix2f *h_camera = &camera;
	Matrix2f *d_camera;
	img_t *d_camera_elements;

	checkCudaErrors(cudaMalloc((void **) &d_camera, sizeof(Matrix2f)));
	checkCudaErrors(cudaMalloc((void **) &(d_camera_elements), sizeof(img_t) * h_camera->width * h_camera->height));
	checkCudaErrors(cudaMemcpy(d_camera_elements, h_camera->elements, sizeof(img_t) * h_camera->width * h_camera->height, cudaMemcpyHostToDevice));
	checkCudaErrors(cudaMemcpy(d_camera, h_camera, sizeof(Matrix2f), cudaMemcpyHostToDevice));
	checkCudaErrors(cudaMemcpy(&(d_camera->elements), &(d_camera_elements), sizeof(img_t*), cudaMemcpyHostToDevice));

	LOG_INFO(logger, __FMT_RUNNING_SIMULATION);
	LOG_INFO(logger, __FMT_SIMULATION_INPUT, "Samples: ", samples, from, to);
	LOG_INFO(output, "%s %s %s", "Probability", "Time", "Deviation");

	LOG4CXX_INFO(logger, kSimulationTableBorders);
	LOG_INFO(logger, __FMT_SIMULATION_TABS, "N", "Skip", "Time (sec)", "Correlation", "(x, y)");
	LOG4CXX_INFO(logger, kSimulationTableBorders);
	for (int p = from_int; p <= to_int; ++p) {
		skip_ = p / 10.0F;

		CorrelationResult summary_res;
		summary_res.deviation = 0;
		summary_res.time = 0;
		int _samples = p != 0 ? samples : 1;

		for (iteration_ = 1; iteration_ <= _samples; ++iteration_) {

			SparseImage pattern = GetSparseImage<unsigned char>(m_pattern_uc, skip_);
			img_t pattern_sum = 0;
			NormalizeImage(pattern, pattern_sum);
			// TODO:
			int x_out = 0;
			int y_out = 0;
			CorrelationResult res = FindCorrelation(d_camera, pattern, pattern_sum, x_out, y_out);

			summary_res.deviation += res.deviation;
			summary_res.time += res.time;

			// Free allocated resources
			pattern.freeResources();
		}
		summary_res.deviation /= _samples;
		summary_res.time /= _samples;

		LOG4CXX_INFO(logger, kSimulationTableBorders);
		LOG_INFO(logger, __FMT_SIMULATION_AVEGAGE_RESULT, "AVR", "TIME", summary_res.time, "DEVIATION", summary_res.deviation);
		LOG4CXX_INFO(logger, kSimulationTableBorders);

		LOG_INFO(output, "%f %f %f", skip_, summary_res.time, summary_res.deviation);
	}

	// Cleanup memory
	cudaFree(d_camera_elements);
	cudaFree(d_camera);

	cudaDeviceReset();
}

// Estimating average time and deviation with fixed probability
void GPUCorrelation2::find(unsigned samples, float skip_probability, int &x_out, int &y_out) {
	// Load camera to device memory
	Matrix2f *h_camera = &camera;
	Matrix2f *d_camera;
	img_t *d_camera_elements;

	checkCudaErrors(cudaMalloc((void **) &d_camera, sizeof(Matrix2f)));
	checkCudaErrors(cudaMalloc((void **) &(d_camera_elements), sizeof(img_t) * h_camera->width * h_camera->height));
	checkCudaErrors(cudaMemcpy(d_camera_elements, h_camera->elements, sizeof(img_t) * h_camera->width * h_camera->height, cudaMemcpyHostToDevice));
	checkCudaErrors(cudaMemcpy(d_camera, h_camera, sizeof(Matrix2f), cudaMemcpyHostToDevice));
	checkCudaErrors(cudaMemcpy(&(d_camera->elements), &(d_camera_elements), sizeof(img_t*), cudaMemcpyHostToDevice));

	LOG_INFO(logger, __FMT_RUNNING_SIMULATION);
	LOG_INFO(logger, __FMT_SIMULATION_INPUT, "Samples: ", samples, skip_probability, skip_probability);
	LOG_INFO(output, "%s %s %s", "Probability", "Time", "Deviation");

	LOG4CXX_INFO(logger, kSimulationTableBorders);
	LOG_INFO(logger, __FMT_SIMULATION_TABS, "N", "Skip", "Time (sec)", "Correlation", "(x, y)");
	LOG4CXX_INFO(logger, kSimulationTableBorders);

	skip_ = skip_probability;

	CorrelationResult summary_res;
	summary_res.deviation = 0;
	summary_res.time = 0;
	int _samples = samples;

	for (iteration_ = 1; iteration_ <= _samples; ++iteration_) {

		SparseImage pattern = GetSparseImage<unsigned char>(m_pattern_uc, skip_);
		img_t pattern_sum = 0;
		NormalizeImage(pattern, pattern_sum);
		CorrelationResult res = FindCorrelation(d_camera, pattern, pattern_sum, x_out, y_out);

		summary_res.deviation += res.deviation;
		summary_res.time += res.time;

		// Free allocated resources
		pattern.freeResources();
	}
	summary_res.deviation /= _samples;
	summary_res.time /= _samples;

	LOG4CXX_INFO(logger, kSimulationTableBorders);
	LOG_INFO(logger, __FMT_SIMULATION_AVEGAGE_RESULT, "AVR", "TIME", summary_res.time, "DEVIATION", summary_res.deviation);
	LOG4CXX_INFO(logger, kSimulationTableBorders);

	LOG_INFO(output, "%f %f %f", skip_, summary_res.time, summary_res.deviation);

	// Cleanup memory
	cudaFree(d_camera_elements);
	cudaFree(d_camera);

	cudaDeviceReset();

}

}	// namespace gpu
}	// namespace recognition
