#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#include <log4cxx/logger.h>
#include <base/logging.h>

#include <base/types.cuh>
#include <base/methods.cuh>
#include <cpu/cpu_correlation.h>

namespace recognition {
namespace cpu {

const log4cxx::LoggerPtr CPUCorrelation::logger = log4cxx::LoggerPtr(log4cxx::Logger::getLogger("recognition.cpu.CPUCorrelation"));
const log4cxx::LoggerPtr CPUCorrelation::output = log4cxx::LoggerPtr(log4cxx::Logger::getLogger("output.recognition.cpu.CPUCorrelation"));

CorrelationResult CPUCorrelation::FindCorrelation(Matrix2f pattern,
		int *stoch_vect, unsigned v_length, img_t pattern_sum, unsigned n_elements) {
	int M = camera.height, N = camera.width;
	int m = pattern.height, n = pattern.width;
	int mn = m * n;

	unsigned stoch_index = 0;

	// start the timer
	clock_t start_time;
	clock_t end_time;
	start_time = clock();

	img_t max_corr = 0;		// maximum found correlation
	int x_res, y_res;		// resulting point (x, y) of best max_corr

	// cycle through the picture
	for (int y = 0; y < M - m + 1; y++) {		// y (rows)
		for (int x = 0; x < N - n + 1; x++) {	// x (columns)
			img_t corr = 0;				// correlation on current iteration
			img_t avr = 0;
			img_t norm = 0;
			int avr_ = 0;				// average of current segment mxn
			long norm_ = 0;				// average of squared values in segment mxn

			if (n_elements < mn) { // sparse
				stoch_index = 0;
				for (int yy = 0; yy < m; ++yy) {
					for (int xx = 0; xx < n; ++xx) {
						if (stoch_vect[stoch_index++ % v_length]) {
							continue;
						}

						img_t val = (img_t) camera.elements[(y + yy) * N + x + xx];
						avr_ += val;
						norm_ += val * val;
					}
				}
			} else {
				// We can use integral representation
				avr_ = EvalIntegralReprSum<int>(camera_integral, N, y, x, m, n);
				norm_ = EvalIntegralReprSum<long>(camera_integral2, N, y, x, m, n);
			}

			avr = (img_t) avr_ / n_elements;
			norm = (img_t) norm_ / n_elements;
			norm -= avr * avr;
			norm = norm != 0 ? sqrt(norm) : 1;

			// Calculate the product
			if (n_elements < mn) { // sparse
				stoch_index = 0;
				for (int yy = 0; yy < m; yy++) {
					for (int xx = 0; xx < n; xx++) {
						// if time to skip, just skip
						if (stoch_vect[stoch_index++ % v_length]) {
							continue;
						}

						corr += pattern.elements[yy * n + xx] * camera.elements[(y + yy) * N + x + xx];
//						cur_F += pattern[yy * n + xx] * ((float)camera[alfa * ((y + yy) * N + x + xx)] - Avr);
//						cur_F += pattern[yy * n + xx] * ((float)camera[alfa * ((y + yy) * N + x + xx)] - Avr) / Norm;	// redundant division
					}
				}
			} else {
				for (int yy = 0; yy < m; yy++) {
					for (int xx = 0; xx < n; xx++) {
						// Don't do redundant if
						corr += pattern.elements[yy * n + xx] * camera.elements[(y + yy) * N + x + xx];
					}
				}
			}

			corr = (corr - avr * pattern_sum) / norm;
//			cur_F /= Norm;
//			corr /= mn;	// redundant division

			if (corr > max_corr) {
				y_res = y;
				x_res = x;
				max_corr = corr;
			}
		}
	}

	end_time = clock();
	CorrelationResult res;
	res.time = (float)(end_time - start_time) / 1000000.0F;
	res.deviation = sqrt(pow(x_res - x_target, 2) + pow(y_res - y_target, 2));

	LOG_INFO(logger, __FMT_SIMULATION_ITERATION_RESULT, iteration_, skip_, res.time, max_corr, x_res, y_res);
	return res;
}

// Estimating average time and deviation with fixed probability
void CPUCorrelation::average(unsigned samples, float from, float to) {
	if (samples <= 0) {
		LOG_ERROR(logger, "average takes positive number of samples");
		exit(EXIT_FAILURE);
	}

	if (from > to || from < 0.0F || to >= 1.0F) {
		LOG_ERROR(logger, "average takes from and to in range [0, 1) and from should be not greater than to");
		exit(EXIT_FAILURE);
	}

	int from_int = (int) (from * 10);
	int to_int = (int) (to * 10);

	LOG_INFO(logger, __FMT_RUNNING_SIMULATION);
	LOG_INFO(logger, __FMT_SIMULATION_INPUT, "Samples: ", samples, from, to);
	LOG_INFO(output, "%s %s %s", "Probability", "Time", "Deviation");

	LOG4CXX_INFO(logger, kSimulationTableBorders);
	LOG_INFO(logger, __FMT_SIMULATION_TABS, "N", "Skip", "Time (sec)", "Correlation", "(x, y)");
	LOG4CXX_INFO(logger, kSimulationTableBorders)
	for (int p = from_int; p <= to_int; ++p) {
		skip_ = p / 10.0F;

		CorrelationResult summary_res;
		summary_res.deviation = 0;
		summary_res.time = 0;
		int _samples = p != 0 ? samples : 1;

		for (iteration_ = 1; iteration_ <= _samples; ++iteration_) {
			// Matrix of pattern with normalized elements of type img_t
			Matrix2f npattern;
			unsigned v_length = 10000;
			int *stoch_vect = (int*) malloc(sizeof(int) * v_length);
			img_t npattern_sum = 0;
			unsigned n_elements = 0;
			npattern = NormalizeImage(m_pattern_uc, stoch_vect, v_length, skip_, npattern_sum, n_elements);

			CorrelationResult res = FindCorrelation(npattern, stoch_vect, v_length, npattern_sum, n_elements);
			free(stoch_vect);
			free(npattern.elements);

			summary_res.deviation += res.deviation;
			summary_res.time += res.time;
		}
		summary_res.deviation /= _samples;
		summary_res.time /= _samples;

		LOG4CXX_INFO(logger, kSimulationTableBorders);
		LOG_INFO(logger, __FMT_SIMULATION_AVEGAGE_RESULT, "AVR", "TIME", summary_res.time, "DEVIATION", summary_res.deviation);
		LOG4CXX_INFO(logger, kSimulationTableBorders);

		LOG_INFO(output, "%f %f %f", skip_, summary_res.time, summary_res.deviation);
	}
}

}	// namespace cpu
}	// namespace recognition
