#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#include <log4cxx/logger.h>
#include <base/logging.h>

#include <base/types.cuh>
#include <base/methods.cuh>
#include <cpu/cpu_correlation2.h>

namespace recognition {
namespace cpu {

const log4cxx::LoggerPtr CPUCorrelation2::logger = log4cxx::LoggerPtr(log4cxx::Logger::getLogger("recognition.cpu.CPUCorrelation2"));
const log4cxx::LoggerPtr CPUCorrelation2::output = log4cxx::LoggerPtr(log4cxx::Logger::getLogger("output.recognition.cpu.CPUCorrelation2"));

// Experiment with the fixed probability
CorrelationResult CPUCorrelation2::FindCorrelation(SparseImage pattern, img_t pattern_sum) {
	int M = camera.height;
	int N = camera.stride;
	int m = pattern.height;
	int n = pattern.stride;

	// start the timer
	clock_t start_time;
	clock_t end_time;
	start_time = clock();

	img_t max_corr = 0;		// maximum found correlation
	int x_res, y_res;		// resulting point (x, y) of best max_corr

	PointList start = pattern.list->ofSelf();
	// Cycle through camera
	for (int y = 0; y < M - m + 1; y++) {		// y (rows)
		for (int x = 0; x < N - n + 1; x++) {	// x (columns)
			img_t corr = 0;				// correlation on current iteration
			img_t avr = 0;				// average of current segment mxn
			img_t norm = 0;				// average of squared values in segment mxn
			int avr_ = 0;				// sum of current segment mxn
			long norm_ = 0;				// sum of squared values in segment mxn

			PointList *list = &start;
			while ((list = list->next) != NULL) {
				int val = camera.elements[(y + list->y) * N + x + list->x];
				corr += pattern.elements[list->y * n + list->x] * val;
				avr_ += val;
				norm_ += val * val;
			}

			avr = (img_t) avr_ / pattern.n_elements;
			norm = (img_t) norm_ / pattern.n_elements;
			norm -= avr * avr;
			norm = norm != 0 ? sqrt(norm) : 1;

			corr = (corr - avr * pattern_sum) / (norm * pattern.n_elements);

			if (corr > max_corr) {
				y_res = y;
				x_res = x;
				max_corr = corr;
			}
		}
	}

	end_time = clock();
	CorrelationResult res;
	res.time = (float) (end_time - start_time) / 1000000.0F;
	res.deviation = sqrt(pow(x_res - x_target, 2) + pow(y_res - y_target, 2));

	LOG_INFO(logger, __FMT_SIMULATION_ITERATION_RESULT, iteration_, skip_, res.time, max_corr, x_res, y_res);
	return res;
}

// Estimating average time and deviation with fixed probability
void CPUCorrelation2::average(unsigned samples, float from, float to) {
	if (samples <= 0) {
		LOG_ERROR(logger, "average takes positive number of samples");
		exit(EXIT_FAILURE);
	}

	if (from > to || from < 0.0F || to >= 1.0F) {
		LOG_ERROR(logger, "average takes from and to in range [0, 1) and from should be not greater than to");
		exit(EXIT_FAILURE);
	}

	int from_int = (int) (from * 10);
	int to_int = (int) (to * 10);

	LOG_INFO(logger, __FMT_RUNNING_SIMULATION);
	LOG_INFO(logger, __FMT_SIMULATION_INPUT, "Samples: ", samples, from, to);
	LOG_INFO(output, "%s %s %s", "Probability", "Time", "Deviation");

	LOG4CXX_INFO(logger, kSimulationTableBorders);
	LOG_INFO(logger, __FMT_SIMULATION_TABS, "N", "Skip", "Time (sec)", "Correlation", "(x, y)");
	LOG4CXX_INFO(logger, kSimulationTableBorders);
	for (int p = from_int; p <= to_int; ++p) {
		skip_ = p / 10.0F;

		CorrelationResult summary_res;
		summary_res.deviation = 0;
		summary_res.time = 0;
		int _samples = p != 0 ? samples : 1;

		for (iteration_ = 1; iteration_ <= _samples; ++iteration_) {
			SparseImage pattern = GetSparseImage<unsigned char>(m_pattern_uc, skip_);
			img_t pattern_sum = 0;
			NormalizeImage(pattern, pattern_sum);
			CorrelationResult res = FindCorrelation(pattern, pattern_sum);

			summary_res.deviation += res.deviation;
			summary_res.time += res.time;

			// Free allocated resources
			pattern.freeResources();
		}
		summary_res.deviation /= _samples;
		summary_res.time /= _samples;

		LOG4CXX_INFO(logger, kSimulationTableBorders);
		LOG_INFO(logger, __FMT_SIMULATION_AVEGAGE_RESULT, "AVR", "TIME", summary_res.time, "DEVIATION", summary_res.deviation);
		LOG4CXX_INFO(logger, kSimulationTableBorders);

		LOG_INFO(output, "%f %f %f", skip_, summary_res.time, summary_res.deviation);
	}
}

}	// namespace cpu
}	// namespace recognition
