#include <stdio.h>
#include <time.h>

#include <log4cxx/logger.h>
#include <base/logging.h>

#include <base/types.cuh>
#include <base/methods.cuh>
#include <gpu/cu_methods.cuh>
#include <gpu/gpu_correlation.cuh>

namespace recognition {
namespace gpu {

const log4cxx::LoggerPtr GPUCorrelation::logger = log4cxx::LoggerPtr(log4cxx::Logger::getLogger("recognition.gpu.GPUCorrelation"));
const log4cxx::LoggerPtr GPUCorrelation::output = log4cxx::LoggerPtr(log4cxx::Logger::getLogger("output.recognition.gpu.GPUCorrelation"));

// Image correlation kernel:
// A - camera, B - pattern, C - each elements of matrix is a correlation of A and B
__global__ void _cu_fullCorrelationKernel(Matrix2f *A, Matrix2f *B, CorrelationPoint *C, _cu_ImgCorrArgs *args) {
	// Block row and column
	int blockRow = blockIdx.y;
	int blockCol = blockIdx.x;

	int row = threadIdx.y;
	int col = threadIdx.x;

	int x = blockCol * BLOCK_SIZE + col;
	int y = blockRow * BLOCK_SIZE + row;

	// Each thread computes one element of Csub (correlation)
	// by accumulating results into Cvalue
	img_t Cvalue = 0;

	int size_B = B->height * B->stride;

	// Each thread block computes one sub-matrix Csub of C
	// Thread row and column are indexes within Csub
//	Matrix2f Csub = _cu_getSubMatrix(C, blockRow, blockCol);

	// Loop over all the sub-matrices of A and B that are required to
	// compute Csub. Find correlation between each pair of sub-matrices
	// and accumulate the results
	for (int m = 0; m < (B->height / BLOCK_SIZE); ++m) {
		for (int n = 0; n < (B->width / BLOCK_SIZE); ++n) {
			// Get sub-matrix Asub of A
			Matrix2f Asub = _cu_getSubMatrix(A, blockRow + m, blockCol + n);

			// Get sub-matrix Bsub of B
			Matrix2f Bsub = _cu_getSubMatrix(B, m, n);

			// Shared memory used to store Asub and Bsub respectively
			__shared__ img_t As[SUBMATRIX_SIZE][SUBMATRIX_SIZE];
			__shared__ img_t Bs[BLOCK_SIZE][BLOCK_SIZE];

			// Load Asub and Bsub from device memory to shared memory.
			// Each thread loads four elements of each sub-matrix as follows:
			//	 _ _ _ _ _
			//	|     |   :
			//	|  a _|a _:
			//	|_ _:_|   :
			//	:  a:  a  :
			//	:_ _:_ _ _:
			//
			// Note, that border elements (row == 0 or col == 0) except the
			// corner element (row == 0 && col == 0) make one redundant read,
			// which is more efficient, than augmenting this loops with
			// conditional read, depending on border case
			for (int i = 0; i < 2; ++i) {
				for (int j = 0; j < 2; ++j) {
					int prow_ = j > 0 ? BLOCK_SIZE - 1 : 0;
					int pcol_ = i > 0 ? BLOCK_SIZE - 1 : 0;
					As[row + prow_][col + pcol_] = _cu_getElement(Asub, row + prow_, col + pcol_);
				}
			}

			Bs[row][col] = _cu_getElement(Bsub, row, col);

			// Synchronize to make sure the sub-matrices are loaded
			// before starting the computation
			__syncthreads();

			// Find partial correlation between Asub and Bsub
			if (args->n_elements < size_B) {
				for (int mm = 0; mm < BLOCK_SIZE; ++mm) {
					for (int nn = 0; nn < BLOCK_SIZE; ++nn) {
						int stoch_index = ((m * BLOCK_SIZE + mm) * B->stride + (n * BLOCK_SIZE + nn)) % args->v_length;
						if (args->d_stoch_vect[stoch_index]) {
							continue;
						}

						Cvalue += As[row + mm][col + nn] * Bs[mm][nn];
					}
				}
			} else {
				for (int mm = 0; mm < BLOCK_SIZE; ++mm) {
					for (int nn = 0; nn < BLOCK_SIZE; ++nn) {
						Cvalue += As[row + mm][col + nn] * Bs[mm][nn];
					}
				}
			}

			// Synchronize to make sure that the preceding computation is done
			// before loading two new sub-matrices of A and B in the next iteration
			__syncthreads();
		}
	}

	int m = B->height;
	int n = B->width;

	img_t avr = 0;
	img_t norm = 0;
	int A_sum = 0;		// sum of elements in current segment (B.width x B.height) of matrix A
	long A_sum2 = 0;	// sum of squared values of elements in segment (B.width x B.height) of matrix A

	if (args->n_elements < size_B) { // sparse
		int stoch_index = 0;
		for (int yy = 0; yy < m; ++yy) {
			for (int xx = 0; xx < n; ++xx) {
				if (args->d_stoch_vect[stoch_index++ % args->v_length]) {
					continue;
				}

				img_t val = (img_t) A->elements[(y + yy) * A->stride + x + xx];
				A_sum += val;
				A_sum2 += val * val;
			}
		}
	} else {
		// We can use integral representation
		A_sum = EvalIntegralReprSum<int>(args->d_integral, A->stride, y, x, m, n);
		A_sum2 = EvalIntegralReprSum<long>(args->d_integral2, A->stride, y, x, m, n);
	}

	avr = (img_t) A_sum / args->n_elements;
	norm = (img_t) A_sum2 / args->n_elements;
	norm -= avr * avr;
	norm = norm != 0 ? sqrt(norm) : 1;

	Cvalue = (Cvalue - avr * args->pattern_sum) / norm;

//	if (x == 356 && y == 137) {
//		printf("\n>>>>>>>(%d, %d) : Cvalue=%f, A[x,y]=%f", x, y, Cvalue, A->elements[y * A->stride + x]);
//	}

	atomicAdd(&(args->debug), 1.0F);

	CorrelationPoint res;
	res.y = y;
	res.x = x;
	res.corr = Cvalue;
	C[y * (A->stride - B->stride) + x] = res;

//	// Write Csub to device memory. Each thread writes one element
//	_cu_setElement(Csub, row, col, res);
}

// TODO: implement border case for x = 386 || y = 386 (x, y start from 0 and end 511)
// Image dimensions are assumed to be multiples of BLOCK_SIZE
CorrelationResult GPUCorrelation::FindCorrelation(Matrix2f *d_camera, Matrix2f pattern, _cu_ImgCorrArgs *d_args) {
    // Load B to device memory
	Matrix2f *h_pattern = &pattern;
	Matrix2f *d_pattern;
	img_t *d_pattern_elements;

	_cu_catch(cudaMalloc((void **) &(d_pattern), sizeof(Matrix2f)), "cudaMalloc d_pattern");

	// Allocate and copy elements
	_cu_catch(cudaMalloc((void **) &(d_pattern_elements), sizeof(img_t) * h_pattern->stride * h_pattern->height),
			"cudaMalloc d_pattern_elements");

	_cu_catch(cudaMemcpy(d_pattern_elements, h_pattern->elements, sizeof(img_t) * h_pattern->stride * h_pattern->height, cudaMemcpyHostToDevice),
			"cudaMemcpy d_pattern_elements h_pattern->elements");

	_cu_catch(cudaMemcpy(d_pattern, h_pattern, sizeof(Matrix2f), cudaMemcpyHostToDevice),
				"cudaMemcpy d_pattern h_pattern");

	_cu_catch(cudaMemcpy(&(d_pattern->elements), &(d_pattern_elements), sizeof(img_t*), cudaMemcpyHostToDevice),
					"cudaMemcpy &(d_pattern->elements) &(d_pattern_elements)");

	// Allocate correlation matrix in device memory
	CorrelationPoint *g_idata;

	_cu_catch(cudaMalloc((void **) &g_idata, sizeof(CorrelationPoint) * (camera.height - h_pattern->height) * (camera.stride - h_pattern->stride)),
			"cudaMalloc g_idata returned");

	// Allocate CUDA events that we'll use for timing
	cudaEvent_t start;
	_cu_catch(cudaEventCreate(&start), "Failed to create start event");

	cudaEvent_t stop;
	_cu_catch(cudaEventCreate(&stop), "Failed to create stop event");

	// Record the start event
	_cu_catch(cudaEventRecord(start, NULL), "Failed to record start event");

	// Find correlation
	dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);
	dim3 dimGrid((camera.stride - h_pattern->stride) / dimBlock.x, (camera.height - h_pattern->height) / dimBlock.y);
	_cu_fullCorrelationKernel<<<dimGrid, dimBlock>>>(d_camera, d_pattern, g_idata, d_args);

	cudaError_t error = cudaGetLastError();

	if (error != cudaSuccess) {
		fprintf(stderr, "error=%s\n", cudaGetErrorString(error));
		exit(EXIT_FAILURE);
	}

	/////////////////// Find maximum correlation in result matrix ////////////////////////
	// Compute max correlation
	int n_threads = N_THREADS_FOR_REDUCE;
	int n_elements = (camera.height - h_pattern->height) * (camera.stride - h_pattern->stride);
	// reduces the number of blocks, therefore burdens threads with more cycle iterations
	int multiplier = 16;
	int n_blocks = n_elements / (n_threads * multiplier);
	int size = n_blocks;

	// Allocate reduce output array
	CorrelationPoint *g_odata;

	_cu_catch(cudaMalloc((void **) &g_odata, sizeof(CorrelationPoint) * size), "cudaMalloc g_odata");

	// Look for maximum correlation
	{
		dim3 dimBlock(n_threads);
		dim3 dimGrid(n_blocks);
		unsigned int smemSize = dimBlock.x;	// should work with dynamic shared mem allocation

		switch (dimBlock.x) {
		case 1024:
			_cu_reduceMaxKernel<1024><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 512:
			_cu_reduceMaxKernel< 512><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 256:
			_cu_reduceMaxKernel< 256><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 128:
			_cu_reduceMaxKernel< 128><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 64:
			_cu_reduceMaxKernel<  64><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 32:
			_cu_reduceMaxKernel<  32><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 16:
			_cu_reduceMaxKernel<  16><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 8:
			_cu_reduceMaxKernel<   8><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 4:
			_cu_reduceMaxKernel<   4><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 2:
			_cu_reduceMaxKernel<   2><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		case 1:
			_cu_reduceMaxKernel<   1><<< dimGrid, dimBlock, smemSize >>>(g_idata, g_odata, n_elements); break;
		}
	}

//	_cu_findMaximum<<<n_blocks, n_threads>>>(g_idata, g_odata);
	CorrelationPoint *data = (CorrelationPoint*) malloc(sizeof(CorrelationPoint) * size);

	_cu_catch(cudaMemcpy(data, g_odata, sizeof(CorrelationPoint) * size, cudaMemcpyDeviceToHost),
			"cudaMemcpy data g_odata");

	CorrelationPoint cPoint;
	cPoint.x = cPoint.y = -1;
	cPoint.corr = 0;

	for (int i = 0; i < size; ++i) {
		CorrelationPoint cur_corr = data[i];
		if (cPoint.corr < cur_corr.corr) {
			cPoint = cur_corr;
		}
	}

	img_t corr = cPoint.corr;
	int y_res = cPoint.y;
	int x_res = cPoint.x;

	free(data);
	//////////////////////////////////////////////////////////////////////////////////////

	 // Record the stop event
	_cu_catch(cudaEventRecord(stop, NULL), "Failed to record stop event");

	// Wait for the stop event to complete
	_cu_catch(cudaEventSynchronize(stop), "Failed to synchronize on the stop event");

	float msecTotal = 0.0f;
	_cu_catch(cudaEventElapsedTime(&msecTotal, start, stop), "Failed to get time elapsed between events");

//	_cu_findCorrelationKernelArgs _debug(0);
//	_cu_findCorrelationKernelArgs *debug = &_debug;
//	_cu_catch(cudaMemcpy(debug, d_args, sizeof(_cu_findCorrelationKernelArgs), cudaMemcpyDeviceToHost),
//			"cudaMemcpy debug d_args");

	// Compute and print the performance
	float sec = msecTotal / 1000.0F;

	cudaFree(g_odata);
	cudaFree(g_idata);
	cudaFree(d_pattern_elements);
	cudaFree(d_pattern);

	CorrelationResult res;
	res.time = sec;
	res.deviation = sqrt(pow(x_res - x_target, 2) + pow(y_res - y_target, 2));

	LOG_INFO(logger, __FMT_SIMULATION_ITERATION_RESULT, iteration_, skip_, sec, corr, x_res, y_res);
	return res;
}

// Estimating average time and deviation with fixed probability
void GPUCorrelation::average(unsigned samples, float from, float to) {
	if (samples <= 0) {
		LOG_ERROR(logger, "average takes positive number of samples");
		exit(EXIT_FAILURE);
	}

	if (from > to || from < 0.0F || to >= 1.0F) {
		LOG_ERROR(logger, "average takes from and to in range [0, 1) and from should be not greater than to");
		exit(EXIT_FAILURE);
	}

	int from_int = (int) (from * 10);
	int to_int = (int) (to * 10);

	// Load camera to device memory
	Matrix2f *h_camera = &camera;
	Matrix2f *d_camera;
	img_t *d_camera_elements;
	int MN = h_camera->stride * h_camera->height;

	_cu_catch(cudaMalloc((void **) &d_camera, sizeof(Matrix2f)), "cudaMalloc d_camera");

	_cu_catch(cudaMalloc((void **) &(d_camera_elements), sizeof(img_t) * MN), "cudaMalloc d_camera_elements");

	_cu_catch(cudaMemcpy(d_camera_elements, h_camera->elements, sizeof(img_t) * MN, cudaMemcpyHostToDevice),
			"cudaMemcpy d_camera_elements h_camera->elements");

	_cu_catch(cudaMemcpy(d_camera, h_camera, sizeof(Matrix2f), cudaMemcpyHostToDevice), "cudaMemcpy d_camera h_camera");

	_cu_catch(cudaMemcpy(&(d_camera->elements), &(d_camera_elements), sizeof(img_t*), cudaMemcpyHostToDevice),
				"cudaMemcpy d_camera->elements d_camera_elements");

	int *d_integral;
	long *d_integral2;

	_cu_catch(cudaMalloc((void **) &(d_integral), sizeof(int) * MN), "cudaMalloc d_integral");

	_cu_catch(cudaMemcpy(d_integral, camera_integral, sizeof(int) * MN, cudaMemcpyHostToDevice),
				"cudaMemcpy d_integral camera_integral");

	_cu_catch(cudaMalloc((void **) &(d_integral2), sizeof(long) * MN), "cudaMalloc d_integral2");

	_cu_catch(cudaMemcpy(d_integral2, camera_integral2, sizeof(long) * MN, cudaMemcpyHostToDevice),
					"cudaMemcpy d_integral2 camera_integral2");

	LOG_INFO(logger, __FMT_RUNNING_SIMULATION);
	LOG_INFO(logger, __FMT_SIMULATION_INPUT, "Samples: ", samples, from, to);
	LOG_INFO(output, "%s %s %s", "Probability", "Time", "Deviation");

	LOG4CXX_INFO(logger, kSimulationTableBorders);
	LOG_INFO(logger, __FMT_SIMULATION_TABS, "N", "Skip", "Time (sec)", "Correlation", "(x, y)");
	LOG4CXX_INFO(logger, kSimulationTableBorders);
	for (int p = from_int; p <= to_int; ++p) {
		skip_ = p / 10.0F;

		CorrelationResult summary_res;
		summary_res.deviation = 0;
		summary_res.time = 0;
		int _samples = p != 0 ? samples : 1;

		for (iteration_ = 1; iteration_ <= _samples; ++iteration_) {
			// Matrix of pattern with normalized elements of type img_t
			Matrix2f npattern;
			unsigned v_length = 10000;
			int *stoch_vect = (int*) malloc(sizeof(int) * v_length);
			img_t npattern_sum = 0;
			unsigned n_elements = 0;
			npattern = NormalizeImage(m_pattern_uc, stoch_vect, v_length, skip_, npattern_sum, n_elements);

			int *d_stoch_vect;
			_cu_catch(cudaMalloc((void **) &(d_stoch_vect), sizeof(int) * v_length), "cudaMalloc d_stoch_vect");

			_cu_catch(cudaMemcpy(d_stoch_vect, stoch_vect,  sizeof(int) * v_length, cudaMemcpyHostToDevice),
							"cudaMemcpy d_stoch_vect stoch_vect");

			_cu_ImgCorrArgs _args(d_integral, d_integral2, d_stoch_vect, v_length, npattern_sum, n_elements);
			_cu_ImgCorrArgs *h_args = &_args;
			_cu_ImgCorrArgs *d_args;

			_cu_catch(cudaMalloc((void **) &d_args, sizeof(_cu_ImgCorrArgs)), "cudaMalloc d_args");

			_cu_catch(cudaMemcpy(d_args, h_args, sizeof(_cu_ImgCorrArgs), cudaMemcpyHostToDevice),
					"cudaMemcpy d_args h_args");


			CorrelationResult res = FindCorrelation(d_camera, npattern, d_args);
			cudaFree(d_args);
			cudaFree(d_stoch_vect);
			free(stoch_vect);
			free(npattern.elements);

			summary_res.deviation += res.deviation;
			summary_res.time += res.time;
		}
		summary_res.deviation /= _samples;
		summary_res.time /= _samples;

		LOG4CXX_INFO(logger, kSimulationTableBorders);
		LOG_INFO(logger, __FMT_SIMULATION_AVEGAGE_RESULT, "AVR", "TIME", summary_res.time, "DEVIATION", summary_res.deviation);
		LOG4CXX_INFO(logger, kSimulationTableBorders);

		LOG_INFO(output, "%f %f %f", skip_, summary_res.time, summary_res.deviation);
	}

	// Free device memory
	cudaFree(d_integral);
	cudaFree(d_integral2);
	cudaFree(d_camera_elements);
	cudaFree(d_camera);

	cudaDeviceReset();
}

}	// namespace gpu
}	// namespace recognition

