#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include <log4cxx/logger.h>
#include <log4cxx/xml/domconfigurator.h>
#include <base/logging.h>

#include <cv.h>
#include <highgui.h>
#include "opencv2/imgproc/imgproc.hpp"
#include "opencv2/core/core.hpp"

#include <base/types.cuh>
#include <recognition.cuh>

#include <camera.h>
#include <video.cuh>
#include <main.cuh>

namespace recognition {

unsigned char* ReadBitmapImageFile(const char *name) {
	FILE *filePtr;						//our file pointer
	BITMAPFILEHEADER bitmapFileHeader;	//our bitmap file header
	unsigned char *bitmapImage;  		//store image data

	filePtr = fopen(name,"rb");
	if (filePtr == NULL)
		return NULL;

	//read the bitmap file header
	fread(&bitmapFileHeader, sizeof(BITMAPFILEHEADER),1,filePtr);
//	printf("type: %x\n size: %x\n offset: %x\n", bitmapFileHeader.bfType, bitmapFileHeader.bfSize, bitmapFileHeader.bOffBits);

	// verify that this is a bmp file by check bitmap id
	if (bitmapFileHeader.bfType !=0x4D42) {
		fprintf(stderr, "Error! The provided file \"%s\" has different from bitmap image type! %i", name, bitmapFileHeader.bfType);
		fclose(filePtr);
		return NULL;
	}

	// Move file pointer to the begging of bitmap data
	fseek(filePtr, bitmapFileHeader.bOffBits, SEEK_SET);

	// Allocate enough memory for the bitmap image data
	bitmapImage = (unsigned char*) malloc(bitmapFileHeader.bfSize - bitmapFileHeader.bOffBits);

	// Read in the bitmap image data
	fread(bitmapImage, bitmapFileHeader.bfSize - bitmapFileHeader.bOffBits, 1, filePtr);

	// Make sure bitmap image data was read
	if (bitmapImage == NULL) {
		fprintf(stderr, "Error reading bitmap image from file \"%s\"", name);
		fclose(filePtr);
		return NULL;
	}

	// Close file and return bitmap image data
	if (fclose(filePtr) != 0) {
		fprintf(stderr, "Error closing the file \"%s\"", name);
	}

	return bitmapImage;
}

unsigned char* RawBmpToRgb(unsigned char *image, int height, int width) {
	unsigned char *res;
	int size = width * height * 3, x, y;
	res = (unsigned char*) malloc(size);

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			int to = ((height - y - 1) * width + x) * 3;
			int from = (y * width + x) * 3;
			res[to] = image[from + 2];
			res[to + 1] = image[from + 1];
			res[to + 2] = image[from];
		}
	}

	return res;
}

unsigned char* ReadBmpRgb(const char *name, int height, int width) {
	unsigned char *bitmapImage, *res;

	bitmapImage = ReadBitmapImageFile(name);
	res = RawBmpToRgb(bitmapImage, height, width);

	free(bitmapImage);

	return res;
}

unsigned char* BmpRgbToGrayscale(unsigned char *image, int height, int width) {
	unsigned char *res;
	int size = width * height, x, y;
	res = (unsigned char*) malloc(size);

	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			int idx = y * width + x;
			int idx3 = idx * 3;
			unsigned char R, G, B, Y;
			R = image[idx3];
			G = image[idx3 + 1];
			B = image[idx3 + 2];

			Y = 0.299 * R + 0.587 * G + 0.114 * B;	// YUV
//			Y = 0.2125 * R + 0.7154 * G + 0.0721 * B; // BT.709

			res[idx] = Y;
//			printf("\n(x, y) = (%d, %d)   R=%d G=%d B=%d    Y=%d", x, y, R, G, B, Y);
		}
	}

	free(image);

	return res;
}

}	// namespace recognition

using namespace std;
using namespace recognition;

// Main Function
int main(int argc, const char * argv[]) {

	// Testing video recognition real time
//	video_test();
//	exit(0);

	// Configuring logging
	string configFileStr(CONFIG_DIRECTORY);
	configFileStr.push_back(FILE_PATH_SEPARATOR);
	configFileStr.append(CONFIG_FILE_XML);
	log4cxx::xml::DOMConfigurator::configure(configFileStr.c_str());

	// Get default logger
	log4cxx::LoggerPtr logger(log4cxx::Logger::getRootLogger());
	LOG_INFO(logger, "Entering application.");
	LOG_INFO(logger, "Reading command line arguments...");

	if (!argv[1] || !argv[2]) {
		LOG_ERROR(logger, "Both input files must be specified!");
		exit(EXIT_FAILURE);
	}

	string imagesDirStr(RESOURCES_DIRECTORY);
	imagesDirStr.push_back(FILE_PATH_SEPARATOR);

	// Input camera and pattern images' file names
	string cameraFileStr(imagesDirStr + argv[1]);
	string patternFileStr(imagesDirStr + argv[2]);

	int M, N;						// Default camera height and width
	int m, n;						// Default pattern height and width

	M = N = 512;
	m = n = 128;

	int x_target, y_target;			// Target (x, y) coordinates of top left corner of pattern

	x_target = y_target = -1;

	// Read the remaining args from command line
	sscanf( argv[3], "%d", &M );
	sscanf( argv[4], "%d", &N );
	sscanf( argv[5], "%d", &m );
	sscanf( argv[6], "%d", &n );
	sscanf( argv[7], "%d", &x_target );
	sscanf( argv[8], "%d", &y_target );

	LOG4CXX_INFO(logger, kInputArgumentsTableBorders);
	LOG_INFO(logger, "|%-15s|%-40s|%-10s|%-10s|", "Image", "File", "Width", "Height");
	LOG4CXX_INFO(logger, kInputArgumentsTableBorders);
	LOG_INFO(logger, "|%-15s|%-40s|%-10d|%-10d|", "camera", cameraFileStr.c_str(), N, M);
	LOG_INFO(logger, "|%-15s|%-40s|%-10d|%-10d|", "pattern", patternFileStr.c_str(), n, m);
	LOG4CXX_INFO(logger, kInputArgumentsTableBorders);
	LOG_INFO(logger, "Target (x, y)  =  (%d, %d)", x_target, y_target);

//	exit(0);

	unsigned char *camera_elements_uc;		// camera image bytes
	unsigned char *pattern_elements_uc;		// pattern image bytes

	// Read the data and convert it into gray scale
	camera_elements_uc = BmpRgbToGrayscale(ReadBmpRgb(cameraFileStr.c_str(), M, N), M, N);
	pattern_elements_uc = BmpRgbToGrayscale(ReadBmpRgb(patternFileStr.c_str(), m, n), m, n);

	// Read images using cv::imread
//	cv::Mat cam = cv::imread(cameraFileStr.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
//	cv::Mat pat = cv::imread(patternFileStr.c_str(), CV_LOAD_IMAGE_GRAYSCALE);
//	camera_elements_uc = cam.ptr(0);
//	pattern_elements_uc = pat.ptr(0);
	/*
		... your code here
	*/

	// Output camera and pattern images' file names
//	string outPrefix("8bit_");
//	string cameraOutFileStr(imagesDirStr + outPrefix + argv[1]);
//	string patternOutFileStr(imagesDirStr + outPrefix + argv[2]);
//
//	cv::imwrite(cameraOutFileStr, cam);
//	cv::imwrite(patternOutFileStr, pat);
//
//	cam.deallocate();
//	pat.deallocate();
//
//	printf("Ending writing new files...\n");
//	exit(0);

	// TODO:
	int x_out = 0;
	int y_out = 0;
	// Run recognition
	Recognition recognition = Recognition::Builder()
		.CameraHeight(M)
		.CameraWidth(N)
		.PatternHeight(m)
		.PatternWidth(n)
		.X_target(x_target)
		.Y_target(y_target)
		.Camera_elements_uc(camera_elements_uc)
		.Pattern_elements_uc(pattern_elements_uc)
		.build();

	recognition.recognize(x_out, y_out);

	// Free images data
	LOG_INFO(logger, "Freeing images data...");
	free(camera_elements_uc);
	free(pattern_elements_uc);

	LOG4CXX_INFO(logger, "Exiting application.");

    return 0;
}
